﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using MathNet.Numerics.LinearAlgebra;

using Agora.Model.Utils;

namespace Agora.Model.ImageAnalysis
{

    public enum CDimensionType
    {
        NoType = 0,
        EUCLIDIAN_X,
        EUCLIDIAN_Y,
        RANGEMAP_X,
        RANGEMAP_Y,
        CHILDS_NUMBER
    }

    abstract public class CDimension
    {
        public CDimensionType Type { get; set; }
        public CDimension(CDimensionType type)
        {
            Type = type;
        }
        public string GetFriendlyName()
        {
            return System.Enum.GetName(typeof(CDimensionType), Type);
        }
        abstract public double GetCoordinate(Vector properties);
    }

    public class CDimensionEuclidian : CDimension
    {
        public double Min, Max;
        public CDimensionEuclidian(CDimensionType type, CEOC eoc)
            : base(type)
        {
            switch (type)
            {
                case CDimensionType.EUCLIDIAN_X:
                    Min = (double)eoc.BoundingBox.Left;
                    Max = (double)eoc.BoundingBox.Right;
                    break;
                case CDimensionType.EUCLIDIAN_Y:
                    Min = (double)eoc.BoundingBox.Top;
                    Max = (double)eoc.BoundingBox.Bottom;
                    break;
            }
        }
        override public double GetCoordinate(Vector properties)
        {

            double result = 0.0;
            switch (Type)
            {
                case CDimensionType.EUCLIDIAN_X: result = properties[0]; break;
                case CDimensionType.EUCLIDIAN_Y: result = properties[1]; break;
            }
            return result;
        }
    }

    public class CDimensionChildsNumber : CDimension
    {
        public CDimensionChildsNumber(CDimensionType type, CEOC eoc)
            : base(type)
        {
        }
        override public double GetCoordinate(Vector properties)
        {
            double result = 0.0;
            return result;
        }
    }

    public class CDimensionRangeMap : CDimension
    {
        private const int BINARIZATION_THRESHOLD = 128;
        private const UInt32 COLOR_WHITE = 0xFFFFFFFF;
        private const UInt32 COLOR_BLACK = 0xFF000000;

        private byte[,] _Map;
        private int _XLength;
        private int _YLength;
        public double Min, Max;

        private void StoreMap(uint[,] map, int xmin, int xmax, int ymin, int ymax)
        {
            Trace.WriteLine("Storing map...");
            for (int y = ymin; y <= ymax; y++)
            {
                for (int x = xmin; x <= xmax; x++)
                {
                    uint mapValue = map[x, y];
                    _Map[x, y] = Convert.ToByte(mapValue);
                }
            }
            Trace.WriteLine("Storing map OK.");
        }

        private void Normalize(uint[,] map, double minValue, double maxValue, int xmin, int xmax, int ymin, int ymax)
        {
            Trace.WriteLine("Computing min/max for normalisation...");
            uint minimum = map[0, 0];
            uint maximum = map[0, 0];
            for (int y = ymin; y <= ymax; y++)
            {
                for (int x = xmin; x <= xmax; x++)
                {
                    minimum = Math.Min(minimum, map[x, y]);
                    maximum = Math.Max(maximum, map[x, y]);
                }
            }
            Trace.WriteLine("Computing min/max OK.");

            Trace.WriteLine("Normalisation between " + minValue.ToString() + " and " + maxValue.ToString() + " ...");
            double a = (maxValue - minValue) / Convert.ToDouble(maximum - minimum);
            double b = (minValue * Convert.ToDouble(maximum) - maxValue * Convert.ToDouble(minimum)) / Convert.ToDouble(maximum - minimum);
            for (int y = ymin; y <= ymax; y++)
            {
                for (int x = xmin; x <= xmax; x++)
                {
                    double currentValue = Convert.ToDouble(map[x, y]);
                    double result = a * currentValue + b;
                    uint mapValue = Convert.ToUInt32(Math.Truncate(result));
                    map[x, y] = mapValue;
                }
            }
            Trace.WriteLine("Normalisation OK.");
        }

        private void Derivate(uint[,] map, bool horizontal, int xmin, int xmax, int ymin, int ymax)
        {

            Trace.WriteLine("Derivation...");
            if (horizontal == true)
            {
                for (int y = 0; y <= ymax; y++)
                {
                    uint previousValue = map[0, y];
                    uint delta;
                    uint currentValue;
                    for (int x = 0; x <= xmax; x++)
                    {
                        currentValue = map[x, y];
                        delta = (uint)Math.Abs((int)currentValue - (int)previousValue);

                        map[x, y] = delta;

                        previousValue = currentValue;
                    }
                }
            }
            else
            {
                for (int x = 0; x <= xmax; x++)
                {
                    uint previousValue = map[x, 0];
                    uint delta;
                    uint currentValue;
                    for (int y = 0; y <= ymax; y++)
                    {
                        currentValue = map[x, y];
                        delta = (uint)Math.Abs((int)currentValue - (int)previousValue);

                        map[x, y] = delta;

                        previousValue = currentValue;
                    }
                }
            }
            Trace.WriteLine("Derivation OK.");

        }
        
        private void Integrate(uint[,] map, bool horizontal, int xmin, int xmax, int ymin, int ymax)
        {
            
            Trace.WriteLine("Integration...");
            if (horizontal == true)
            {
                for (int y = 0; y <= ymax; y++)
                {
                    uint previousValue = map[0, y];
                    uint currentValue;
                    for (int x = 0; x <= xmax; x++)
                    {
                        currentValue = map[x, y];

                        map[x, y] = currentValue + previousValue;

                        previousValue = previousValue + currentValue;
                    }
                }
            }
            else
            {
                for (int x = 0; x <= xmax; x++)
                {
                    uint previousValue = map[x, 0];
                    uint currentValue;
                    for (int y = 0; y <= ymax; y++)
                    {
                        currentValue = map[x, y];

                        map[x, y] = currentValue + previousValue;

                        previousValue = previousValue + currentValue;
                    }
                }
            }
            Trace.WriteLine("Integration OK.");
            
        }

        private void ComputeRanges(uint[,] map, bool[,] whitePixel, bool horizontal, int xmin, int xmax, int ymin, int ymax)
        {
            Trace.WriteLine("Computing white consecutive pixels...");
            if (horizontal == true)
            {
                for (int x = 0; x <= xmax; x++)
                {
                    int y = 0;
                    while (y <= ymax)
                    {
                        //UInt32 argbColor;
                        uint nbWhitePixels;
                        int yy = y;
                        while ((yy <= ymax) && (whitePixel[x, yy] == true))
                            //((argbColor = (UInt32)image.GetPixel(x, yy).ToArgb()) == COLOR_WHITE))
                        {
                            yy++;
                        }

                        nbWhitePixels = (uint)(yy - y);

                        do
                        {
                            map[x, y] = nbWhitePixels;
                            y++;
                        }
                        while (y < yy);

                    }
                }
            }
            else
            {
                for (int y = 0; y <= ymax; y++)
                {
                    int x = 0;
                    while (x <= xmax)
                    {
                        //UInt32 argbColor;
                        uint nbWhitePixels;
                        int xx = x;
                        while ((xx <= xmax) && (whitePixel[xx, y] == true))
                            //((argbColor = (UInt32)image.GetPixel(xx, y).ToArgb()) == COLOR_WHITE))
                        {
                            xx++;
                        }

                        nbWhitePixels = (uint)(xx - x);

                        do
                        {
                            map[x, y] = nbWhitePixels;
                            x++;
                        }
                        while (x < xx);
                    }
                }
            }
            Trace.WriteLine("Computing OK.");
        }

        private void ExtractWhitePixels(Bitmap binImage, bool[,] whitePixel, int xmin, int xmax, int ymin, int ymax) 
        {
            Trace.WriteLine("Extract white pixels...");
            Bitmap tmpBinImage = binImage.Clone(
                new Rectangle(0, 0, binImage.Width, binImage.Height),
                PixelFormat.Format32bppArgb);
            FastBitmap tmpImage = new FastBitmap(tmpBinImage);
            tmpImage.LockImage();
            for (int y = ymin; y <= ymax; y++)
            {
                for (int x = xmin; x <= xmax; x++)
                {
                    /*
                    UInt32 argbColor = (UInt32)binImage.GetPixel(x, y).ToArgb();
                    if (argbColor == COLOR_WHITE)
                    */
                    //Color c = tmpImage.GetPixel(x, y);
                    FastBitmap.PixelData32bpp pix = tmpImage.GetPixel32bpp(x, y);
                    if ((pix.red == 255) && (pix.green == 255) && (pix.blue == 255))
                    //if ((c.B == 255) && (c.G == 255) && (c.R == 255))
                        whitePixel[x, y] = true;
                    else
                        whitePixel[x, y] = false;
                }
            }
            tmpImage.UnlockImage();
            Trace.WriteLine("Extract white pixels OK.");
        }

        private void ComputeMap(CEOC eoc)
        {

            string name = new StackTrace().GetFrame(0).GetMethod().ReflectedType.Name;
            Trace.WriteLine(Type.ToString() + "...", name);

            Bitmap tmpImage;
            FastBitmap fastBitmap;
            int gray;
            //Color c;
            int xmin = eoc.BoundingBox.Left;
            int xmax = eoc.BoundingBox.Right - 1;
            int ymin = eoc.BoundingBox.Top;
            int ymax = eoc.BoundingBox.Bottom - 1;
            _XLength = eoc.BoundingBox.Width;
            _YLength = eoc.BoundingBox.Height;
            _Map = new byte[_XLength, _YLength];
            uint[,] tmpMap = new uint[_XLength, _YLength];

            FastBitmap fastBmp = new FastBitmap(eoc.GetImage());
            fastBmp.LockImage();
            byte[] bitmapHashCode = fastBmp.GetHashCode();
            fastBmp.UnlockImage();
            fastBmp = null;
            Guid guid = new Guid(bitmapHashCode);

            string id = name + "." + Type.ToString() + "." + guid.ToString();
            if (!CImageAnalysis.Cache.Exist(id))
            {
                tmpImage = eoc.GetImage().Clone(eoc.BoundingBox, eoc.GetImage().PixelFormat);
                //tmpImage.Save("map_debug00.jpg");

                Trace.WriteLine("Grayscale...");
                Bitmap grayImage = AForge.Imaging.Filters.Grayscale.CommonAlgorithms.BT709.Apply(tmpImage);
                Trace.WriteLine("Grayscale OK.");

                Trace.WriteLine("Contrast...");
                AForge.Imaging.Filters.ContrastStretch equFilter = new AForge.Imaging.Filters.ContrastStretch();
                equFilter.ApplyInPlace(grayImage);
                Trace.WriteLine("Contrast OK.");

                Trace.WriteLine("Noise reduction...");
                AForge.Imaging.Filters.Median medianFilter = new AForge.Imaging.Filters.Median(3);
                //medianFilter.ApplyInPlace(grayImage);
                Trace.WriteLine("Noise reduction OK.");

                //CImageAnalysis.Cache.StoreInCache("map_debug00", grayImage);

                Trace.WriteLine("Binarization " + BINARIZATION_THRESHOLD.ToString() + " ...");
                AForge.Imaging.Filters.Threshold thresholdFilter =
                    new AForge.Imaging.Filters.Threshold(BINARIZATION_THRESHOLD);
                //AForge.Imaging.Filters.SauvolaThreshold sauvolaFilter = new AForge.Imaging.Filters.SauvolaThreshold(12, 0.5F, 128.0F);
                Bitmap binImage = grayImage; // sauvolaFilter.Apply(grayImage);
                thresholdFilter.ApplyInPlace(binImage);
                Trace.WriteLine("Binarization OK.");

                //CImageAnalysis.Cache.StoreInCache("map_debug01", binImage);

                Trace.WriteLine("Erosion, dilatation, closing, opening...");
                // BEWARE !!! AForge's "Erosion" is Dilatation !!!
                AForge.Imaging.Filters.Erosion3x3 erosionFilter = new AForge.Imaging.Filters.Erosion3x3();
                AForge.Imaging.Filters.Dilatation3x3 dilatationFilter = new AForge.Imaging.Filters.Dilatation3x3();
                AForge.Imaging.Filters.Closing closingFilter = new AForge.Imaging.Filters.Closing();
                AForge.Imaging.Filters.Opening openingFilter = new AForge.Imaging.Filters.Opening();
                //dilatationFilter.ApplyInPlace(binImage);
                //erosionFilter.ApplyInPlace(binImage);
                //closingFilter.ApplyInPlace(binImage);
                //openingFilter.ApplyInPlace(binImage);
                Trace.WriteLine("Erosion, etc. OK.");

                //CImageAnalysis.Cache.StoreInCache("map_debug02", binImage);

                Trace.WriteLine("Compute range map...");
                bool[,] whitePixels = new bool[_XLength, _YLength];
                this.ExtractWhitePixels(binImage, whitePixels, xmin, xmax, ymin, ymax);
                switch (Type)
                {
                    case CDimensionType.RANGEMAP_X:
                        this.ComputeRanges(tmpMap, whitePixels, true, xmin, xmax, ymin, ymax);
                        //this.Normalize(tmpMap, 0.0, 4.0, xmin, xmax, ymin, ymax);
                        //this.Derivate(tmpMap, true, xmin, xmax, ymin, ymax);
                        //this.Normalize(tmpMap, 0.0, 2.0, xmin, xmax, ymin, ymax);
                        //this.Integrate(tmpMap, true, xmin, xmax, ymin, ymax);
                        //this.Normalize(tmpMap, 0.0, 10.0, xmin, xmax, ymin, ymax);
                        break;
                    case CDimensionType.RANGEMAP_Y:
                        this.ComputeRanges(tmpMap, whitePixels, false, xmin, xmax, ymin, ymax);
                        //this.Normalize(tmpMap, 0.0, 4.0, xmin, xmax, ymin, ymax);
                        //this.Derivate(tmpMap, false, xmin, xmax, ymin, ymax);
                        //this.Normalize(tmpMap, 0.0, 2.0, xmin, xmax, ymin, ymax);
                        //this.Integrate(tmpMap, false, xmin, xmax, ymin, ymax);
                        //this.Normalize(tmpMap, 0.0, 10.0, xmin, xmax, ymin, ymax);
                        break;
                }
                Trace.WriteLine("Compute range map OK.");

                this.Normalize(tmpMap, Min, Max, xmin, xmax, ymin, ymax);
                this.StoreMap(tmpMap, xmin, xmax, ymin, ymax);

                Trace.WriteLine("Compute map image for cache storing...");
                this.Normalize(tmpMap, 0.0, 255.0, xmin, xmax, ymin, ymax);

                Bitmap newBitmap = new Bitmap(_XLength, _YLength, PixelFormat.Format32bppArgb);
                fastBitmap = new FastBitmap(newBitmap);
                fastBitmap.LockImage();
                for (int y = ymin; y <= ymax; y++)
                {
                    for (int x = xmin; x <= xmax; x++)
                    {
                        gray = (int)tmpMap[x, y];
                        /*
                        c = Color.FromArgb(gray, gray, gray);
                        tmpImage.SetPixel(x, y, c);
                        */
                        FastBitmap.PixelData32bpp pix;
                        pix.red = pix.green = pix.blue = (byte)gray;
                        pix.alpha = 255;
                        //c = Color.FromArgb(255, gray, gray, gray);
                        fastBitmap.SetPixel32bpp(x, y, pix);
                    }
                }
                fastBitmap.UnlockImage();
                CImageAnalysis.Cache.StoreInCache(id, newBitmap);
                Trace.WriteLine("Compute map image for cache storing OK."); 
                
                Trace.WriteLine("Garbage collect to free huge temporary memory...");
                tmpMap = null;
                System.GC.Collect();
                Trace.WriteLine("Garbage collect OK.");
            }
            else
            {
                Trace.WriteLine("Loading map image from cache ...");
                tmpImage = CImageAnalysis.Cache.LoadFromCache(id);
                Trace.WriteLine("Loading map image from cache OK.");
                Trace.WriteLine("Building map image from loaded map image ...");
                fastBitmap = new FastBitmap(tmpImage);
                fastBitmap.LockImage();
                for (int y = ymin; y <= ymax; y++)
                {
                    for (int x = xmin; x <= xmax; x++)
                    {
                        /*
                        c = tmpImage.GetPixel(x, y);
                        gray = (int)c.R; // ((int)c.R + (int)c.G + (int)c.B) / 3;
                        */
                        FastBitmap.PixelData32bpp pix = fastBitmap.GetPixel32bpp(x, y);
                        gray = pix.red;
                        //c = fastBitmap.GetPixel(x, y);
                        //gray = c.R;
                        tmpMap[x, y] = (uint)gray;
                    }
                }
                fastBitmap.UnlockImage();
                this.Normalize(tmpMap, Min, Max, xmin, xmax, ymin, ymax);
                this.StoreMap(tmpMap, xmin, xmax, ymin, ymax);
                Trace.WriteLine("Building map image from loaded map image OK.");

                // for debug:
                /*
                this.Normalize(tmpMap, 0.0, 255.0, xmin, xmax, ymin, ymax);
                for (int y = ymin; y <= ymax; y++)
                {
                    for (int x = xmin; x <= xmax; x++)
                    {
                        gray = (int)tmpMap[x, y];
                        c = Color.FromArgb(gray, gray, gray);
                        tmpImage.SetPixel(x, y, c);
                    }
                }
                tmpImage.Save("debug001.jpg");
                */
            }

            Trace.WriteLine("OK.", name); 
        }

        public CDimensionRangeMap(CDimensionType type, double min, double max, CEOC eoc)
            : base(type)
        {
            Min = min; Max = max;
            this.ComputeMap(eoc);
        }

        override public double GetCoordinate(Vector properties)
        {
            int xmap = Convert.ToInt32(Math.Truncate(properties[0]));
            int ymap = Convert.ToInt32(Math.Truncate(properties[1]));
            double result;
            if ((xmap >= 0) && (xmap < _XLength) && (ymap >= 0) && (ymap < _YLength))
            {
                result = Convert.ToDouble(_Map[xmap, ymap]);
            }
            else
            {
                result = Double.NaN;
            }
            return result;
        }
    }

}
