﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Drawing;

using MathNet.Numerics.LinearAlgebra;

using Agora.Model.Utils;

namespace Agora.Model.ImageAnalysis
{

    public class CEOCType
    {
        // Predefined EOC types
        public const int EOCNoType   = 0; public const string EOCNoTypeName   = "NONAME";
        public const int EOCDocument = 1; public const string EOCDocumentName = "DOCUMENT";
        public const int EOCCC       = 2; public const string EOCCCName       = "CC";
        public const int EOCNoise    = 3; public const string EOCNoiseName    = "NOISE";
        public const int EOCChar     = 4; public const string EOCCharName     = "CHAR";
        public const int EOCImage    = 5; public const string EOCImageName    = "IMAGE";
        public const int EOCWord     = 6; public const string EOCWordName     = "WORD";
        public const int EOCLine     = 7; public const string EOCLineName     = "LINE";
        public const int EOCBlock    = 8; public const string EOCBlockName    = "BLOCK";
        public const int EOCVector   = 9; public const string EOCVectorName   = "VECTOR";
        public const int EOCTmp      = 10; public const string EOCTmpName     = "TMP";

        public int Type;
        public string TypeName;

        public CEOCType()
        {
            Type = CEOCType.EOCNoType;
            TypeName = null;
        }
        public CEOCType(int id, string s)
        {
            Type = id;
            TypeName = s;
        }
        /*public CEOCType(int t)
        {
            Type = t;
            TypeName = null; // new string(t.ToString());
        }*/
        public override string ToString()
        {
            return "{" + Type.ToString() + "," + TypeName + "}";
        }
    }

    public class CEOC
    {
        /*
        private CGraph _Graph = null;
        public CGraph Graph
        {
            get { return _Graph; }
        }
        */
        private CEOC _Parent = null;
        public CEOC Parent
        {
            get { return _Parent; }
        }
        
        private CEOCType _Eoctype = null;
        public CEOCType Eoctype
        {
            get { return _Eoctype; }
            set { _Eoctype = value; }
        }

        private LinkedListNode<CEOC> _Node = null;
        public LinkedListNode<CEOC> Node
        {
            get { return _Node; }
            set { _Node = value; }
        }

        private LinkedList<CEOC> _Childs = new LinkedList<CEOC>();
        public LinkedList<CEOC> Childs
        {
            get { return _Childs; }
        }
        
        private LinkedListNode<CEOC> _PairedWith = null;
        public LinkedListNode<CEOC> PairedWith
        {
            get { return _PairedWith; }
            set { _PairedWith = value; }
        }
        private LinkedListNode<CEOC> _PairedWithCandidate = null;
        public LinkedListNode<CEOC> PairedWithCandidate
        {
            get { return _PairedWithCandidate; }
            set { _PairedWithCandidate = value; }
        }

        private CFeatures _Features = null;
        public CFeatures Features
        {
            get { return _Features; }
            set { _Features = value; }
        }

        private CFeatureMode[] _FeatureModes = null;
        public CFeatureMode[] FeatureModes
        {
            get { return _FeatureModes; }
            set { _FeatureModes = value; }
        }

        /*
        /// <summary>
        /// Only used for EOC in patterns
        /// </summary>
        private CFeatures _PatternFeatures = null;
        public CFeatures PatternFeatures
        {
            get { return _PatternFeatures; }
            set { _PatternFeatures = value; }
        }
        */

        private CDimension[] _Dimensions = null;
        public CDimension[] Dimensions
        {
            get { return _Dimensions; }
            set { _Dimensions = value; }
        }

        private System.Drawing.Rectangle _BoundingBox = new System.Drawing.Rectangle();
        public System.Drawing.Rectangle BoundingBox
        {
            get { return _BoundingBox; }
            set { _BoundingBox = value; }
        }

        private System.Drawing.Bitmap _Image = null;
        public System.Drawing.Bitmap Image
        {
            get { return _Image; }
            set { _Image = value; }
        }

        public CEOC()
        {
        }
        /*
        public CEOC(CGraph g)
        {
            _Graph = g;
            _Image = null;
            _BoundingBox = new System.Drawing.Rectangle();
            _Eoctype = null;
        }
        */
        public CEOC(/*CGraph g,*/ CEOCType t)
        {
            //_Graph = g;
            //_Image = null;
            //_BoundingBox = new System.Drawing.Rectangle();
            _Eoctype = t;
        }

        public CEOC(/*CGraph g,*/ System.Drawing.Bitmap im, System.Drawing.Rectangle r, CEOCType t)
        {
            //_Graph = g;
            _Image = im;
            _BoundingBox = new System.Drawing.Rectangle(r.X, r.Y, r.Width, r.Height);
            _Eoctype = t;
        }

        public CEOC Clone()
        {
            CEOC result = new CEOC();
            result._BoundingBox = new System.Drawing.Rectangle(
                this._BoundingBox.X, this._BoundingBox.Y,
                this._BoundingBox.Width, this._BoundingBox.Height);
            foreach (CEOC child in _Childs)
            {
                CEOC newChild = child.Clone();
                result.AddChild(newChild);
            }            
            result._Dimensions = this._Dimensions;
            result._Eoctype = this._Eoctype; // will be updated later
            result._Image = this._Image;
            result._Features = this._Features.Clone();
            result._FeatureModes = CFeatureMode.CloneArray(this._FeatureModes);
            return result;
        }
        
        // Needed to maintain relation between EOC
        // and his node in the child list -> easy to remove
        public void AddChild(CEOC child)
        {
            child._Node = new LinkedListNode<CEOC>(child);
            this.AddChild(child._Node);
        }
        public void AddChild(LinkedListNode<CEOC> child)
        {
            child.Value._Parent = this;
            //child.Value._Dimensions = this._Dimensions;
            _Childs.AddLast(child);
        }        
        public void RemoveChild(CEOC child)
        {
            this.RemoveChild(child._Node);
            child._Node = null;
        }
        public void RemoveChild(LinkedListNode<CEOC> child)
        {
            _Childs.Remove(child);
            child.Value._Dimensions = null;
            child.Value._Parent = null;
        }

        public bool IsMatching(CEOC eoc)
        {
            if (eoc == null)
                return false;
            else
                return this.IsMatching(eoc.Eoctype);
        }

        public bool IsMatching(CEOCType type)
        {
            return
                (this._Eoctype.Type == CEOCType.EOCNoType) ||
                (type.Type == CEOCType.EOCNoType) ||
                this._Eoctype.Type == type.Type;
        }

        public void PairWithCandidate(LinkedListNode<CEOC> eoc)
        {
            this._PairedWithCandidate = eoc;
            eoc.Value._PairedWithCandidate = this._Node;
        }
        public void UnpairCandidate()
        {
            LinkedListNode<CEOC> eoc = this._PairedWithCandidate;
            if (eoc != null)
            {
                this._PairedWithCandidate = null;
                eoc.Value._PairedWithCandidate = null;
            }
        }

        public void PairingStore()
        {
            LinkedListNode<CEOC> eoc = this._PairedWithCandidate;
            this.PairingStore(eoc);
        }
        public void PairingStore(LinkedListNode<CEOC> eoc)
        {
            this._PairedWith = eoc;
            eoc.Value._PairedWith = this._Node;
        }
        public void PairingRemove()
        {
            LinkedListNode<CEOC> eoc = this._PairedWith;
            if (eoc != null)
            {
                this._PairedWith = null;
                eoc.Value._PairedWith = null;
            }
        }

        public void ComputeBoundingBox()
        {
            bool first = true;
            Trace.Assert(this._Childs.Count > 0);
            foreach (CEOC e in this._Childs)
            {
                if (first == true)
                {
                    _BoundingBox = System.Drawing.Rectangle.Union(
                        e._BoundingBox,
                        e._BoundingBox);
                    first = false;
                }
                else
                {
                    _BoundingBox = System.Drawing.Rectangle.Union(
                        _BoundingBox,
                        e._BoundingBox);
                }
            }
        }

        public void SetDimensions(CDimensionType[] dimensionTypes)
        {
            _Dimensions = new CDimension[dimensionTypes.Length];

            for (int i = 0; i < dimensionTypes.Length; i++)
            {
                CDimensionType type = dimensionTypes[i];
                switch (type)
                {
                    case CDimensionType.EUCLIDIAN_X:
                    case CDimensionType.EUCLIDIAN_Y:
                        _Dimensions[i] = new CDimensionEuclidian(type, this);
                        break;
                    case CDimensionType.RANGEMAP_X:
                    case CDimensionType.RANGEMAP_Y:
                        _Dimensions[i] = new CDimensionRangeMap(
                            type,
                            GLOBALDATA.RANGEMAP_NORMALIZE_MIN,
                            GLOBALDATA.RANGEMAP_NORMALIZE_MAX,
                            this);
                        break;
                    case CDimensionType.CHILDS_NUMBER:
                        _Dimensions[i] = new CDimensionChildsNumber(type, this);
                        break;
                }
            }
        }
        /// <summary>
        /// Find nearest parent which owns dimension list
        /// </summary>
        /// <returns>Dimension list or null</returns>
        public CDimension[] GetDimensions()
        {
            CDimension[] result = null;
            CEOC eoc = this;
            while ((result = eoc._Dimensions) == null)
            {
                eoc = eoc._Parent;
                if (eoc == null)
                    break;
            }
            return result;
        }

        /// <summary>
        /// Find nearest parent which owns image
        /// </summary>
        /// <returns>Image or null</returns>
        public System.Drawing.Bitmap GetImage()
        {
            System.Drawing.Bitmap result = null;
            CEOC eoc = this;
            while ((result = eoc._Image) == null)
            {
                eoc = eoc._Parent;
                if (eoc == null)
                    break;
            }
            return result;
        }

        private double ComputeZoneAverage(
            Rectangle zone,
            CDimension dimension
            )
        {
            double result = 0.0;
            double nb = 0.0;
            int xmin = zone.Left;
            int xmax = zone.Right - 1;
            int ymin = zone.Top;
            int ymax = zone.Bottom - 1;
            double sum = 0.0; double val;
            Vector props = new Vector((int)2);
            for (int y = ymin; y <= ymax; y++)
            {
                for (int x = xmin; x <= xmax; x++)
                {
                    props[0] = (double)x;
                    props[1] = (double)y;
                    val = dimension.GetCoordinate(props);
                    if (! Double.IsNaN(val))
                    {
                        sum += val;
                        nb += 1.0;
                    }
                }
            }
            if (nb > 0.0)
            {
                result = sum / nb;
            }
            else
            {
                result = Double.NaN;
            }
            return result;
        }

        // Get original features for ACTUAL EOC (ie only physical EOC like CC, vectors, etc.)
        // Other EOC (virtual EOC ?) will obtain their features using FeatureModes
        public void Measure(int eocType)
        {
            CDimension[] dimensions = this.GetDimensions();
            _FeatureModes = null;
            switch (eocType)
            {
                case CEOCType.EOCDocument:
                    // Creates vectors with 0.0 as coordinates
                    _Features = new CFeatures(dimensions.Length);
                    _Features.a[0] = 0.0;
                    _Features.a[1] = 0.0;
                    _Features.b[0] = (double)_BoundingBox.Right - 1.0;
                    _Features.b[1] = (double)_BoundingBox.Bottom - 1.0;
                    break;
                case CEOCType.EOCCC:
                    _Features = new CFeatures(dimensions.Length);
                    // Updates coordinates:
                    for (int numDim = 0; numDim < dimensions.Length; numDim++)
                    {
                        switch (dimensions[numDim].Type)
                        {
                            case CDimensionType.EUCLIDIAN_X:
                                _Features.a[numDim] = (double)_BoundingBox.Left;
                                _Features.b[numDim] = (double)_BoundingBox.Right - 1.0;
                                break;
                            case CDimensionType.EUCLIDIAN_Y:
                                _Features.a[numDim] = (double)_BoundingBox.Top;
                                _Features.b[numDim] = (double)_BoundingBox.Bottom - 1.0;
                                break;
                            case CDimensionType.RANGEMAP_X:
                                // Compute average of left and right zone white range pixels (using RangeMap)
                                double leftAverage = this.ComputeZoneAverage(
                                    new Rectangle(
                                        _BoundingBox.Left - GLOBALDATA.RANGEMAP_LOOK_X,
                                        _BoundingBox.Top,
                                        GLOBALDATA.RANGEMAP_LOOK_X,
                                        _BoundingBox.Height),
                                    dimensions[2]);
                                double rightAverage = this.ComputeZoneAverage(
                                    new Rectangle(
                                        _BoundingBox.Right,
                                        _BoundingBox.Top,
                                        GLOBALDATA.RANGEMAP_LOOK_X,
                                        _BoundingBox.Height),
                                    dimensions[2]);
                                _Features.a[numDim] = Double.IsNaN(leftAverage) ? 0.0 : -leftAverage;
                                _Features.b[numDim] = Double.IsNaN(rightAverage) ? 0.0 : rightAverage;
                                break;
                            case CDimensionType.RANGEMAP_Y:
                                // Compute average of top and bottom zone white range pixels (using RangeMap)
                                double topAverage = this.ComputeZoneAverage(
                                    new Rectangle(
                                        _BoundingBox.Left,
                                        _BoundingBox.Top - GLOBALDATA.RANGEMAP_LOOK_Y,
                                        _BoundingBox.Width,
                                        GLOBALDATA.RANGEMAP_LOOK_Y),
                                    dimensions[3]);
                                double bottomAverage = this.ComputeZoneAverage(
                                    new Rectangle(
                                        _BoundingBox.Left,
                                        _BoundingBox.Bottom,
                                        _BoundingBox.Width,
                                        GLOBALDATA.RANGEMAP_LOOK_Y),
                                    dimensions[3]);
                                _Features.a[numDim] = Double.IsNaN(topAverage) ? 0.0 : -topAverage;
                                _Features.b[numDim] = Double.IsNaN(bottomAverage) ? 0.0 : bottomAverage;
                                break;
                            case CDimensionType.CHILDS_NUMBER:
                                _Features.a[numDim] = 0.0;
                                _Features.b[numDim] = 0.0;
                                break;
                        }
                    }
                    break;
            }
        }

        /// <summary>
        /// Compute Features using internal FeatureMode list
        /// and propagate through parent (call to parent's ComputeFeatures)
        /// </summary>
        /// <returns></returns>
        public void ComputeFeatures()
        {
            // Cannot compute features if no feature modes (ie : EOC = actual EOC)
            // In this case, Features stay as they are.
            if (this._FeatureModes == null)
                return;

            int nbDim = this.GetDimensions().Length;
            if (nbDim != this.FeatureModes.Length)
                throw new Exception("nbDim != this.FeatureModes.Length");

            CFeatures[] childsFeatures = new CFeatures[this.Childs.Count];
            int i = 0;
            double weight = 0.0;
            foreach (CEOC child in this.Childs)
            {
                childsFeatures[i] = child.Features;
                weight += child.Features.weight;
                i++;
            }
            if (weight == 0.0)
                weight = 1.0;
            
            CFeatures newFeatures = new CFeatures(
                weight,
                new double[nbDim],
                new double[nbDim]);

            for (int numFeat = 0; numFeat < nbDim; numFeat++)
            {
                CFeatureMode mode = this.FeatureModes[numFeat];
                switch (GLOBALDATA.DIMENSIONS[numFeat])
                {
                    case CDimensionType.CHILDS_NUMBER:
                        newFeatures.a[numFeat] = this.ComputeFeature(mode.aMode, this.Childs.Count);
                        newFeatures.b[numFeat] = this.ComputeFeature(mode.bMode, this.Childs.Count);
                        break;
                    default:
                        newFeatures.a[numFeat] = this.ComputeFeature(numFeat, true, mode.aMode, childsFeatures);
                        newFeatures.b[numFeat] = this.ComputeFeature(numFeat, false, mode.bMode, childsFeatures);
                        break;
                }
            }

            Features = newFeatures;
            if (((Features.a[4] + Features.b[4]) / 2.0) != this.Childs.Count)
            {
                return;
            }
            Trace.Assert(this != _Parent);
            _Parent.ComputeFeatures();
        }

        private double ComputeFeature(int numFeature, bool a, CFeatureMode.SetFeatureMode mode, CFeatures[] featuresList)
        {
            double val = 0.0;
            double nbValues = 0.0;
            switch (mode)
            {
                case CFeatureMode.SetFeatureMode.Min:
                    val = Double.PositiveInfinity;
                    foreach (CFeatures f in featuresList)
                    {
                        val = Math.Min(val, (a == true) ? f.a[numFeature] : f.b[numFeature]);
                    }
                    break;
                case CFeatureMode.SetFeatureMode.Max:
                    val = Double.NegativeInfinity;
                    foreach (CFeatures f in featuresList)
                    {
                        val = Math.Max(val, (a == true) ? f.a[numFeature] : f.b[numFeature]);
                    }
                    break;
                case CFeatureMode.SetFeatureMode.Average:
                    nbValues = 0.0;
                    val = 0.0;
                    for (int i = 0; i < featuresList.Length; i++)
                    {
                        CFeatures f = featuresList[i];
                        double w = f.weight;
                        double v = (a == true) ? f.a[numFeature] : f.b[numFeature];
                        if (w > 0.0)
                        {
                            val += v * w;
                            nbValues += w;
                        }
                        else
                        {
                            val += v;
                            nbValues += 1.0;
                        }
                    }
                    /*
                    foreach (CFeatures f in featuresList)
                    {
                        val += (a == true) ? f.a[numFeature] : f.b[numFeature];
                        nbValues += 1.0;
                    }
                    */

                    if (nbValues > 0.0)
                        val /= nbValues;
                    break;
                case CFeatureMode.SetFeatureMode.AverageAll:
                    nbValues = 0.0;
                    val = 0.0;
                    for (int i = 0; i < featuresList.Length; i++)
                    {
                        CFeatures f = featuresList[i];
                        double w = f.weight;
                        double v = f.a[numFeature] + f.b[numFeature];
                        if (w > 0.0)
                        {
                            val += v * w;
                            nbValues += 2.0 * w;
                        }
                        else
                        {
                            val += v;
                            nbValues += 2.0;
                        }
                    }
                    /*
                    foreach (CFeatures f in featuresList)
                    {
                        val += f.a[numFeature] + f.b[numFeature];
                        nbValues += 2.0;
                    }
                    */

                    if (nbValues > 0.0)
                        val /= nbValues;
                    break;
            }
            return val;
        }

        private double ComputeFeature(CFeatureMode.SetFeatureMode mode, double n)
        {
            double val = 0.0;
            switch (mode)
            {
                case CFeatureMode.SetFeatureMode.Min:
                    val = n - 0.5;
                    break;
                case CFeatureMode.SetFeatureMode.Max:
                    val = n + 0.5;
                    break;
                case CFeatureMode.SetFeatureMode.Average:
                case CFeatureMode.SetFeatureMode.AverageAll:
                    val = n;
                    break;
            }
            return val;
        }
    }
}
