﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using MathNet.Numerics.LinearAlgebra;
using Agora.Model.Utils;

namespace Agora.Model.ImageAnalysis
{
    public class CFeatureMode
    {
        /// <summary>
        /// Computation of features given list of EOC
        /// </summary>
        public enum SetFeatureMode
        {
            No, Average, Min, Max, AverageAll
        }
        public SetFeatureMode aMode { get; set;}
        public SetFeatureMode bMode { get; set;}

        /// <summary>
        /// Default feature mode : a=Min, b=Max
        /// </summary>
        public CFeatureMode()
        {
            this.aMode = SetFeatureMode.Min;
            this.bMode = SetFeatureMode.Max;
        }
        public CFeatureMode(SetFeatureMode amode, SetFeatureMode bmode)
        {
            this.aMode = amode; this.bMode = bmode;
        }
        public CFeatureMode Clone()
        {
            return new CFeatureMode(this.aMode, this.bMode);
        }
        static public bool CompareArrays(CFeatureMode[] array1, CFeatureMode[] array2)
        {
            if (array1 == null) return false;
            if (array2 == null) return false;
            if (array1.Length != array2.Length) return false;

            bool result = true;
            for (int i = 0; i < array1.Length; i++)
            {
                if (array1[i].aMode != array2[i].aMode) return false;
                if (array1[i].bMode != array2[i].bMode) return false;
            }
            return result;
        }
        static public CFeatureMode[] CloneArray(CFeatureMode[] array)
        {
            if (array == null) return null;
            CFeatureMode[] newArray = new CFeatureMode[array.Length];
            for (int i = 0; i < array.Length; i++)
            {
                if (array[i] != null)
                {
                    newArray[i] = array[i].Clone();
                }
                else
                {
                    newArray[i] = new CFeatureMode();
                }
            }
            return newArray;
        }
        public static SetFeatureMode FromString(string s)
        {
            SetFeatureMode m = SetFeatureMode.No;
            foreach (SetFeatureMode mode in Enum.GetValues(typeof(SetFeatureMode)))
            {
                if (s.CompareTo(Enum.GetName(typeof(SetFeatureMode), mode)) == 0)
                {
                    m = mode;
                }
            }
            /*
            if (s.CompareTo(SetFeatureMode.Average.ToString()) == 0) m = SetFeatureMode.Average;
            if (s.CompareTo(SetFeatureMode.Min.ToString()) == 0) m = SetFeatureMode.Min;
            if (s.CompareTo(SetFeatureMode.Max.ToString()) == 0) m = SetFeatureMode.Max;
            */
            return m;
        }
        public override string ToString()
        {
            string s = "{a=" + System.Enum.GetName(typeof(SetFeatureMode), this.aMode) + ",";
            s += "b=" + System.Enum.GetName(typeof(SetFeatureMode), this.bMode) + "}";
            return s;
        }
        /// <summary>
        /// Builds an array of nb default FeatureMode
        /// </summary>
        /// <param name="nb"></param>
        /// <returns></returns>
        static public CFeatureMode[] BuildArray(int nb)
        {
            CFeatureMode[] a = new CFeatureMode[nb];
            for (int i = 0; i < nb; i++)
            {
                a[i] = new CFeatureMode();
            }
            return a;
        }
    }

    public class CFeatures
    {
        private double _weight = 1.0;
        private double[] _a = null;
        private double[] _b = null;
        public double weight
        {
            get { return _weight; }
            set { _weight = value; }
        }
        public double[] a
        {
            get { return _a; }
            set { _a = value; }
        }
        public double[] b
        {
            get { return _b; }
            set { _b = value; }
        }
        public CFeatures()
        {
        }
        /// <summary>
        /// Default values (all 0.0)
        /// </summary>
        /// <param name="nbDim"></param>
        public CFeatures(int nbDim)
        {
            _a = new double[nbDim];
            _b = new double[nbDim];
        }
        public CFeatures(double[] v0, double[] v1)
        {
            _a = v0;
            _b = v1;
        }
        public CFeatures(double w, double[] v0, double[] v1)
        {
            _weight = w;
            _a = v0;
            _b = v1;
        }
        public CFeatures(Vector v0, Vector v1)
        {
            _a = v0;// uses implicit operator from Vector class
            _b = v1;// uses implicit operator from Vector class
        }
        public CFeatures Clone()
        {
            CFeatures result = new CFeatures(
                this.weight,
                (double[])this.a.Clone(),
                (double[])this.b.Clone());
            return result;
        }
        static public CFeatures[] CloneArray(CFeatures[] array)
        {
            CFeatures[] result = null;
            if (array != null)
            {
                result = new CFeatures[array.Length];
                for (int i = 0; i < array.Length; i++)
                {
                    if (array[i] == null) throw new Exception("Can't clone null Feature.");
                    result[i] = array[i].Clone();
                }
            }
            return result;
        }
        public override string ToString()
        {
            string sa = "[";
            string sb = "[";
            for (int i = 0; i < _a.Length; i++)
            {
                double val = _a[i];
                string valstr = val.ToString("F1");
                sa +=  valstr + ((i == _a.Length-1) ? "" : "|");
                val = _b[i];
                valstr = val.ToString("F1");
                sb += valstr + ((i == _b.Length - 1) ? "" : "|");
            }
            sa += "]";
            sb += "]";
            string s = "{" + "w=" + _weight.ToString() + "," + sa + "," + sb + "}";
            return s;
        }

        /// <summary>
        /// Builds an array of default features
        /// </summary>
        /// <param name="nbFeatures"></param>
        /// <param name="nbDimensions"></param>
        /// <returns></returns>
        public static CFeatures[] BuildArray(int nbFeatures, int nbDimensions)
        {
            CFeatures[] result = new CFeatures[nbFeatures];
            for (int i = 0; i < nbFeatures; i++)
            {
                result[i] = new CFeatures(nbDimensions);
            }
            return result;
        }
    }
}
