﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using MathNet.Numerics.LinearAlgebra;

using Agora.Model.OperatorLevel0.Pairing;
using Agora.Model.ImageAnalysis;
using Agora.Model.Utils;

namespace Agora.Model.OperatorLevel0
{
    public abstract class COperator
    {
        protected int _OperationsCount = 0;
        protected int _RecurseLevel = 0;

        public bool Multiple = true;

        private CPattern _Pattern = null;
        public CPattern Pattern
        {
            get { return _Pattern; }
            set { _Pattern = value; }
        }

        private CFeatureMode[] _FeatureModes = null;
        public CFeatureMode[] FeatureModes
        {
            get { return _FeatureModes; }
            set { _FeatureModes = value; }
        }

        private CPairingEvaluator _PairingEvaluator = null;
        public CPairingEvaluator PairingEvaluator
        {
            get { return _PairingEvaluator; }
            set { _PairingEvaluator = value; }
        }

        private bool _EnableNeighbourhood = false;
        public bool EnableNeighbourhood
        {
            get { return _EnableNeighbourhood; }
            set { _EnableNeighbourhood = value; }
        }
        private double[] _Neighbourhood = null;
        public double[] Neighbourhood
        {
            get { return _Neighbourhood; }
            set { _Neighbourhood = value; }
        }
        private CNeighbourMap _NeighbourMap = null;
        public CNeighbourMap NeighbourMap
        {
            get { return _NeighbourMap; }
            set { _NeighbourMap = value; }
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="p">Pattern</param>
        public COperator(CPattern p)
        {
            _Pattern = p;
        }
        
        /// <summary>
        /// Takes the current pairing (stored in pattern's CANDIDATE pairing)
        /// and evaluates the quality of it
        /// </summary>
        /// <returns>true : pairing is good</returns>
        protected abstract bool EvaluatePairing();

        /// <summary>
        /// Takes the current pairing (stored in pattern's ACTUAL pairing)
        /// and updates the graph according to it
        /// </summary>
        protected abstract void UpdateGraph();

        /// <summary>
        /// Recursively explore possible pairings between Pattern and root.
        /// If a pairing is found (based on EOC's type) Then
        ///     Evaluate this candidate pairing
        ///     If it has a good score, Then
        ///         Save this pairing as "Actual pairing"
        ///         Update graph
        /// Endif
        /// Remark : sometimes, for optimisation needs, the actual pairing
        /// used by UpdateGraph() may differ from the candidate pairing stored
        /// in the pattern. The reason is that we often compute the next candidate
        /// pairing BEFORE graph's update.
        /// </summary>
        /// <param name="root"></param>
        /// <returns>true = at least one graph update has been done</returns>
        public virtual bool Apply(CEOC root)
        {
            //_RecurseLevel++;
            bool result = false;

            //if (result == false)
            //{
            CEOC rootPattern = _Pattern.Graph.GetRoot();
            CEOC eocGraph = root;
            //CEOC eocChild;
            bool processNode = true;
            while (eocGraph != null)
            {
                // Go to deepest left node from current node:
                /*
                if (eocGraph.Childs.First != null)
                    eocChild = eocGraph.Childs.First.Value;
                else
                    eocChild = null;
                while (eocChild != null)
                {
                    eocGraph = eocChild;
                    if (eocChild.Childs.First != null)
                        eocChild = eocChild.Childs.First.Value;
                    else
                        eocChild = null;
                }
                */

                if ((processNode == true) && ((rootPattern.IsMatching(eocGraph) == true)))
                {
                    rootPattern.PairWithCandidate(eocGraph.Node);
                    if (rootPattern.Childs.Count == 0)
                    {
                        // Pattern with no child : pairing is 100% done, so update graph
                        // if pairing has good score
                        if (this.EvaluatePairing() == true)
                        {
                            rootPattern.PairingStore();
                            this.UpdateGraph();
                            rootPattern.PairingRemove();
                            result = true;

                            this.IncOperationsCount();

                        }
                    }
                    else if (rootPattern.Childs.Count == 1)
                    {
                        // Optimisation for single child patterns (more efficient code
                        // than for N childs patterns using arrangement generator)
                        LinkedListNode<CEOC> childPattern = rootPattern.Childs.First;
                        LinkedListNode<CEOC> childGraph = eocGraph.Childs.First;

                        // Try to pair with each child
                        while ((childGraph != null) &&
                            ((result == false) || ((result == true) && (Multiple == true))))
                        {
                            if (childPattern.Value.IsMatching(childGraph.Value) == true)
                            {
                                // set pairing
                                childPattern.Value.PairWithCandidate(childGraph);

                                bool pairingIsGood = this.EvaluatePairing();

                                // get next child before losing childGraph
                                // by updating graph ;) (for example by operator "delete")
                                childGraph = childGraph.Next;

                                if (pairingIsGood == true)
                                {
                                    rootPattern.PairingStore();
                                    childPattern.Value.PairingStore();
                                    this.UpdateGraph();
                                    childPattern.Value.PairingRemove();
                                    rootPattern.PairingRemove();
                                    result = true;

                                    this.IncOperationsCount();

                                }

                                // unset pairing
                                childPattern.Value.UnpairCandidate();
                            }
                            else
                            {
                                childGraph = childGraph.Next;
                            }
                        }
                    }
                    else
                    {
                        // General case : N childs pairing

                        // 1.eventually builds neighbourhood map of eocGraph's childs
                        if (_EnableNeighbourhood == true)
                        {
                            _NeighbourMap = new CNeighbourMap(eocGraph.GetDimensions().Length);
                            _NeighbourMap.Builds(eocGraph, _Neighbourhood);
                        }
                        else
                        {
                            _NeighbourMap = null;
                        }

                        // 2. Explores arrangements until no arrangement produces match

                        CArrangement arrGenerator = new CArrangement(
                            rootPattern, eocGraph, _NeighbourMap);
                        bool arrangementFound = arrGenerator.GetFirst();
                        while ((arrangementFound == true) &&
                            ((result == false) || ((result == true) && (Multiple == true))))

                        {
                            // Pairing was found : evaluate it
                            bool pairingIsGood = this.EvaluatePairing();

                            if (pairingIsGood == true)
                            {
                                // Gets next pairing BEFORE graph modification.
                                // Hopefully this next pairing won't used
                                // currently paired nodes because current pairing is not forgotten
                                // Thank's to a call to PairingStore() we first stores
                                // the current pairing for future graph update
                                arrGenerator.PairingStore();
                                arrangementFound = arrGenerator.GetFirst();
                                this.UpdateGraph();
                                arrGenerator.PairingRemove();
                                result = true;

                                this.IncOperationsCount();

                            }
                            else
                            {
                                arrangementFound = arrGenerator.GetNext();
                            }
                        }
                    }
                    rootPattern.UnpairCandidate();

                }

                // get next node to process
                if ((eocGraph.Childs.First != null) && (processNode == true))
                {
                    // Node has at least one child and node is just processed
                    // Next node = first child of current node
                    eocGraph = eocGraph.Childs.First.Value;
                }
                else if (eocGraph.Node.Next != null)
                {
                    // No childs or node previously processed : next node = next brother
                    eocGraph = eocGraph.Node.Next.Value;
                    processNode = true;
                }
                else
                {
                    // No child, no next brother, or node previously processed :
                    // next node = parent, and disable processing of parent as it is now fully processed
                    eocGraph = eocGraph.Parent;
                    processNode = false;
                }


            }

            /*
            if ((result == false) || ((result == true) && (Multiple == true)))
            {
                LinkedListNode<CEOC> w = root.Childs.First;
                while ((w != null) &&
                    ((result == false) || ((result == true) && (Multiple == true))))
                {
                    result |= this.Apply(w.Value);
                    w = w.Next;
                }
            }
            if ((_OperationsCount > 0) && (_RecurseLevel == 1))
                if ((_OperationsCount % 100) == 0)
                    Trace.WriteLine(_OperationsCount.ToString());
            _RecurseLevel--;
            */

            //Trace.WriteLine("Result = " + _OperationsCount.ToString());
            return result;
        }
        private void IncOperationsCount()
        {
            _OperationsCount++;
            if ((_OperationsCount % 100) == 0)
            {
                Trace.Write("#");
            }
            if ((_OperationsCount % 5000) == 0)
            {
                Trace.WriteLine("#");
            }
        }
    }
}
