﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Xml.Serialization;

using MathNet.Numerics.LinearAlgebra;

using Agora.Model.Utils;
using Agora.Model.ImageAnalysis;

namespace Agora.Model.OperatorLevel0.Pairing
{
    [XmlInclude(typeof(OperatorLevel0.Pairing.CPairingEvaluator))]
    [XmlInclude(typeof(OperatorLevel0.Pairing.CPairingEvaluatorIntersection))]
    public class CPairingEvaluator
    {
        private bool _EnableRotation;
        public bool EnableRotation
        {
            get { return _EnableRotation; }
            set { _EnableRotation = value; }
        }

        private bool _EnableScaling;
        public bool EnableScaling
        {
            get { return _EnableScaling; }
            set { _EnableScaling = value; }
        }

        private bool _EnableTranslation;
        public bool EnableTranslation
        {
            get { return _EnableTranslation; }
            set { _EnableTranslation = value; }
        }

        /// <summary>
        /// N projection matrix 
        /// dim0 = nb of proj (N)
        /// dim1 = coefficients (= nb of dim, only diagonal values)
        /// </summary>
        protected double[][] _Coefficients = null;
        public double[][] Coefficients
        {
            get { return _Coefficients; }
            set { _Coefficients = value; }
        }

        /// <summary>
        /// Comparators between 2 childs (pattern/graph)
        /// dim0 = nb of childs
        /// </summary>
        private CFeaturesComparator[] _PatternGraphComparators = null;
        public CFeaturesComparator[] PatternGraphComparators
        {
            get { return _PatternGraphComparators; }
            set { _PatternGraphComparators = value; }
        }

        /// <summary>
        /// Comparators between each child and his parent
        /// dim0 = nb of childs
        /// </summary>
        private CFeaturesComparator[] _ParentGraphComparators = null;
        public CFeaturesComparator[] ParentGraphComparators
        {
            get { return _ParentGraphComparators; }
            set { _ParentGraphComparators = value; }
        }

        /// <summary>
        /// Comparators between childs (graph/graph)
        /// dim0/1 = nb of childs
        /// </summary>
        private CFeaturesComparator[][] _GraphGraphComparators = null;
        public CFeaturesComparator[][] GraphGraphComparators 
        {
            get { return _GraphGraphComparators; }
            set { _GraphGraphComparators = value; }
        }
        
        protected Matrix[] _Projections = null;

        private CLinearTransformation _Transformation = new CLinearTransformation();

        public CPairingEvaluator()
        {
            _EnableTranslation = false;
            _EnableScaling = false;
            _EnableRotation = false;
        }

        public CPairingEvaluator(
            bool enablerotation, bool enablescaling, bool enabletranslation,
            double[][] coefficients,
            CFeaturesComparator[] patterngraphcomparators,
            CFeaturesComparator[] parentgraphcomparators,
            CFeaturesComparator[][] graphgraphcomparators)
        {
            _EnableRotation = enablerotation;
            _EnableScaling = enablescaling;
            _EnableTranslation = enabletranslation;
            _Coefficients = CPairingEvaluator.CloneArrayArray(coefficients);
            _PatternGraphComparators = CFeaturesComparator.CloneArray(patterngraphcomparators);
            _ParentGraphComparators = CFeaturesComparator.CloneArray(parentgraphcomparators);
            _GraphGraphComparators = CFeaturesComparator.CloneArrayArray(graphgraphcomparators);
        }

        public virtual CPairingEvaluator Clone()
        {
            CPairingEvaluator result = new CPairingEvaluator(
                this._EnableRotation, this._EnableScaling, this._EnableTranslation,
                this._Coefficients,
                this._PatternGraphComparators, this._ParentGraphComparators, this._GraphGraphComparators);
            return result;
        }

        static public double[][] CloneArrayArray(double[][] array)
        {
            double[][] result = null;
            if (array != null)
            {
                result = new double[array.Length][];
                for (int i = 0; i < array.Length; i++)
                {
                    result[i] = (double[])array[i].Clone();
                }
            }
            return result;
        }

        static public double[][] CreateStandardProjection(int nbDim)
        {
            double[][] result = new double[1][];
            result[0] = new double[nbDim];
            for (int i = 0; i < nbDim; i++)
            {
                result[0][i] = 1.0;
            }
            return result;
        }

        static public double[][] CreateNullProjection(int nbDim)
        {
            double[][] result = new double[1][];
            result[0] = new double[nbDim];
            for (int i = 0; i < nbDim; i++)
            {
                result[0][i] = 0.0;
            }
            return result;
        }

        protected void SetProjectionMatrixsFromDiagonalValues(double[][] diagonalValues)
        {
            if (_Projections == null)
            {
                if (diagonalValues != null)
                {
                    this._Projections = new Matrix[diagonalValues.Length];
                    for (int numProj = 0; numProj < diagonalValues.Length; numProj++)
                    {
                        int nbDim = diagonalValues[numProj].Length;
                        Matrix m = Matrix.Diagonal(new Vector(diagonalValues[numProj]));
                        this._Projections[numProj] = m;
                    }
                }
            }
        }

        /// <summary>
        /// Updates list of features from a list of points 
        /// </summary>
        /// <param name="pattern"></param>
        /// <param name="points"></param>
        private void UpdateFeaturesFromPoints(List<CFeatures> features, List<Vector> points)
        {
            int numPoint = 0;
            foreach (CFeatures f in features)
            {
                Vector v0 = points[numPoint];
                Vector v1 = points[numPoint + 1];
                numPoint += 2;
                f.a = v0;
                f.b = v1;
            }
        }

        private List<Vector> GetPointsFromFeatures(List<CFeatures> features)
        {
            List<Vector> points = new List<Vector>();
            foreach (CFeatures f in features)
            {
                points.Add(new Vector(f.a));
                points.Add(new Vector(f.b));
            }
            return points;
        }

        private void GetPairingFeatures(
            CPattern pattern,
            List<CFeatures> patternFeatures,
            List<CFeatures> graphFeatures
            )
        {
            CEOC rootPattern = pattern.Graph.GetRoot();
            foreach (CEOC childPattern in rootPattern.Childs)
            {
                CEOC childGraph = childPattern.PairedWithCandidate.Value;
                Trace.Assert(childPattern.Features != null);
                Trace.Assert(childGraph.Features != null);
                CFeatures f0 = childPattern.Features.Clone();
                patternFeatures.Add(f0);
                CFeatures f1 = childGraph.Features.Clone();
                graphFeatures.Add(f1);
            }
        }

        /// <summary>
        /// Evaluate pairing stored in PairedWith links of pattern.
        /// 1. Gathers points x0,y0 from features of pattern and graph
        /// 2. For each projection :
        ///    a. Computes best transformation A,B for y0=A.x0+B
        ///    b. Builds estimated features from y0e=A.x0+B
        ///       and checks if FeaturesComparatorsPatternGraph are satisfied
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        private bool CheckPairingPatternGraph(CPattern pattern)
        {
            bool allScoresOK = true;

            this.SetProjectionMatrixsFromDiagonalValues(Coefficients);

            for (int numProj = 0; numProj < _Projections.Length; numProj++)
            {
                List<CFeatures> patternFeatures = new List<CFeatures>();
                List<CFeatures> graphFeatures = new List<CFeatures>();
                this.GetPairingFeatures(pattern, patternFeatures, graphFeatures);

                /*
                List<Vector> x0 = this.GetPointsFromFeatures(patternFeatures);
                List<Vector> y0 = this.GetPointsFromFeatures(graphFeatures);
                */

                List<Vector> x0 = this.GetPointsFromFeatures(graphFeatures);
                List<Vector> y0 = this.GetPointsFromFeatures(patternFeatures);

                // Apply projection
                int nbDim = x0[0].Length;
                CLinearTransformation projection = new CLinearTransformation(nbDim);
                projection.A = _Projections[numProj];
                x0 = projection.TransformPoints(x0);
                y0 = projection.TransformPoints(y0);

                _Transformation.ComputeBestTransformation(
                    x0, y0,
                    _EnableTranslation,
                    _EnableScaling,
                    _EnableRotation);
            
                List<Vector> y0e = _Transformation.TransformPoints(x0);

                /*
                this.UpdateFeaturesFromPoints(patternFeatures, y0e);
                this.UpdateFeaturesFromPoints(graphFeatures, y0);
                */

                this.UpdateFeaturesFromPoints(graphFeatures, y0e);
                this.UpdateFeaturesFromPoints(patternFeatures, y0);

                for (int i = 0; i < _PatternGraphComparators.Length; i++)
                {
                    CFeatures f0 = patternFeatures[i];
                    CFeatures f1 = graphFeatures[i];
                    allScoresOK &= _PatternGraphComparators[i].Compare(f0, f1);
                    if (allScoresOK == false) break;
                }
                if (allScoresOK == false) break;
            }
            return allScoresOK;
        }

        /// <summary>
        /// Checks if paired graph's EOCs satisfy FeaturesComparatorsParentGraph
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        private bool CheckPairingParentGraph(CPattern pattern)
        {
            bool allScoresOK = true;
            CEOC rootPattern = pattern.Graph.GetRoot();
            int comparatorNum0 = 0;
            foreach (CEOC childPattern in rootPattern.Childs)
            {
                CEOC childGraph = childPattern.PairedWithCandidate.Value;
                CEOC parentGraph = childGraph.Parent;
                Trace.Assert(rootPattern.PairedWithCandidate.Value == parentGraph);

                CFeatures f0 = childGraph.Features;
                CFeatures f1 = parentGraph.Features;
                CFeaturesComparator fc = _ParentGraphComparators[comparatorNum0];
                allScoresOK &= fc.Compare(f0, f1);
                if (allScoresOK == false) break;
                comparatorNum0++;
            }
            return allScoresOK;
        }

        /// <summary>
        /// Checks if paired graph's EOCs satisfy FeaturesComparatorsGraphGraph
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        private bool CheckPairingGraphGraph(CPattern pattern)
        {
            bool allScoresOK = true;
            CEOC rootPattern = pattern.Graph.GetRoot();
            int comparatorNum0 = 0;
            foreach (CEOC childPattern0 in rootPattern.Childs)
            {
                int comparatorNum1 = 0;
                foreach (CEOC childPattern1 in rootPattern.Childs)
                {
                    CEOC childGraph0 = childPattern0.PairedWithCandidate.Value;
                    CEOC childGraph1 = childPattern1.PairedWithCandidate.Value;
                    CFeatures f0 = childGraph0.Features;
                    CFeatures f1 = childGraph1.Features;
                    CFeaturesComparator fc = _GraphGraphComparators[comparatorNum0][comparatorNum1];
                    allScoresOK &= fc.Compare(f0, f1);
                    if (allScoresOK == false) break;
                    comparatorNum1++;
                }
                if (allScoresOK == false) break;
                comparatorNum0++;
            }
            return allScoresOK;
        }
        public virtual bool EvaluatePairing(CPattern pattern)
        {
            bool result = true;
            if (_PatternGraphComparators != null)
                result &= this.CheckPairingPatternGraph(pattern);
            if (_ParentGraphComparators != null)
                result &= this.CheckPairingParentGraph(pattern);
            if (_GraphGraphComparators != null)
                result &= this.CheckPairingGraphGraph(pattern);
            return result;
        }
    }

    public class CPairingEvaluatorIntersection : CPairingEvaluator
    {

        public CPairingEvaluatorIntersection()
            : base()
        {
        }
        
        public CPairingEvaluatorIntersection(double[][] coefficients)
            : base()
        {
            _Coefficients = CPairingEvaluator.CloneArrayArray(coefficients);
        }

        public override CPairingEvaluator Clone()
        {
            CPairingEvaluatorIntersection result = new CPairingEvaluatorIntersection(
                this._Coefficients);
            return result;
        }

        /// <summary>
        /// Check intersection on all the dimensions. Returns true
        /// if EOCs intersects on all the dimensions (ie EOC instersects)
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public override bool EvaluatePairing(CPattern pattern)
        {
            bool allScoresOK = true;

            this.SetProjectionMatrixsFromDiagonalValues(_Coefficients);

            for (int numProj = 0; numProj < _Projections.Length; numProj++)
            {
                int nbDim = Coefficients[numProj].Length;
                CLinearTransformation projection = new CLinearTransformation(nbDim);
                projection.A = _Projections[numProj];

                CEOC rootPattern = pattern.Graph.GetRoot();
                int comparatorNum0 = 0;
                foreach (CEOC childPattern0 in rootPattern.Childs)
                {
                    int comparatorNum1 = 0;
                    foreach (CEOC childPattern1 in rootPattern.Childs)
                    {
                        CEOC childGraph0 = childPattern0.PairedWithCandidate.Value;
                        CEOC childGraph1 = childPattern1.PairedWithCandidate.Value;

                        if (childGraph0 == childGraph1)
                            continue;

                        // Apply projection
                        List<Vector> points = new List<Vector>();
                        points.Add(new Vector(childGraph0.Features.a));
                        points.Add(new Vector(childGraph0.Features.b));
                        points.Add(new Vector(childGraph1.Features.a));
                        points.Add(new Vector(childGraph1.Features.b));
                        points = projection.TransformPoints(points);

                        CFeatures f0 = new CFeatures(points[0], points[1]);
                        CFeatures f1 = new CFeatures(points[2], points[3]);

                        for (int numFeat = 0; numFeat < f0.a.Length; numFeat++)
                        {
                            double a0 = Math.Min(f0.a[numFeat], f0.b[numFeat]);
                            double b0 = Math.Max(f0.a[numFeat], f0.b[numFeat]);
                            double a1 = Math.Min(f1.a[numFeat], f1.b[numFeat]);
                            double b1 = Math.Max(f1.a[numFeat], f1.b[numFeat]);

                            // Check if NOT intersect !

                            if ((b0 < a1) || (b1 < a0))
                            {
                                allScoresOK = false;
                                break;
                            }
                        }
                        if (allScoresOK == false) break;
                        comparatorNum1++;
                    }
                    if (allScoresOK == false) break;
                    comparatorNum0++;
                }
                if (allScoresOK == false) break;
            }
            return allScoresOK;
        }

        public override string ToString()
        {

            string coefs = "{";
            if (Coefficients != null)
            {
                foreach (double[] coef in Coefficients)
                {
                    coefs += "{";
                    foreach (double d in coef)
                    {
                        coefs = coefs + d.ToString() + ",";
                    }
                    coefs += "},";
                }
            }
            coefs += "}";

            return "EvaluatorIntersection={" +
                "Coefficients=" + coefs + "," +
                "}";
            ;
        }
    }
}
