﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Agora.Model.ImageAnalysis;
using Agora.Model.OperatorLevel0;
using Agora.Model.OperatorLevel0.Pairing;
using Agora.Model.OperatorLevel1;

namespace Agora.Model.OperatorLevel2
{
    public class CClassifyByFeatures : Agora.Model.Project.CStepOperator
    {
        public enum ClassificationType
        {
            No, ClassifyByPosition, ClassifyBySize, ClassifyByForm
        }

        /// <summary>
        /// Context
        /// </summary>
        public string EOCTypeParent { get; set; }

        /// <summary>
        /// Type of the only one child
        /// </summary>
        public string EOCTypeChild { get; set; }

        /// <summary>
        /// Type of the new child if classification has to be done
        /// </summary>
        public string EOCTypeNewEOC { get; set; }

        /// <summary>
        /// Pattern child's features
        /// </summary>
        public CFeatures ChildFeatures { get; set; }

        /// <summary>
        /// N projection matrix 
        /// dim0 = nb of proj (N)
        /// dim1 = coefficients (= nb of dim, only diagonal values)
        /// </summary>
        public double[][] Coefficients { get; set; }

        /// <summary>
        /// Comparator between actual child and pattern child
        /// </summary>
        public CFeaturesComparator Comparator { get; set; }

        /// <summary>
        /// Type of features' comparison
        /// </summary>
        public ClassificationType Type { get; set; }

        /// <summary>
        /// If new EOC is inserted, then replace child EOC
        /// instead of just insert it. This will call "Delete"
        /// </summary>
        public bool Replace { get; set; }

        public CClassifyByFeatures()
        {
            EOCTypeChild = null;
            EOCTypeNewEOC = null;
            EOCTypeParent = null;
            ChildFeatures = null;
            Coefficients = null;
            Comparator = null;
            Type = ClassificationType.ClassifyByPosition;
            Replace = false;
        }

        public CClassifyByFeatures(
            string eocTypeParent, string eocTypeChild, string eocTypeNewEOC,
            CFeatures childFeatures, double[][] coefficients, CFeaturesComparator comparator,
            ClassificationType type, bool replace
            )
        {
            EOCTypeParent = String.Copy(eocTypeParent);
            EOCTypeChild = String.Copy(eocTypeChild);
            EOCTypeNewEOC = String.Copy(eocTypeNewEOC);
            ChildFeatures = childFeatures.Clone();
            Coefficients = CPairingEvaluator.CloneArrayArray(coefficients);
            Comparator = comparator.Clone();
            Type = type;
            Replace = replace;
        }
        public override Agora.Model.Project.CStepOperator Clone()
        {
            return new CClassifyByFeatures(
                EOCTypeParent, EOCTypeChild, EOCTypeNewEOC,
                ChildFeatures, Coefficients, Comparator, Type, Replace);
        }
        public override bool Apply(CImageAnalysis analysis)
        {
            string name = new StackTrace().GetFrame(0).GetMethod().ReflectedType.Name;
            bool result = false;
            
            if (Trace.IndentLevel == 0) Trace.WriteLine("Classify...", name);
            Trace.Indent();

            int nbDim = analysis.GetDimensions().Length;

            OperatorLevel1.CInsert opIns;

            opIns = new CInsert(
                EOCTypeParent, new string[] { EOCTypeChild }, EOCTypeNewEOC,
                new CFeatures[] { ChildFeatures },
                new CPairingEvaluator(
                    false,
                    (Type == ClassificationType.ClassifyByForm) ?
                        true : false,
                    ((Type == ClassificationType.ClassifyByForm) ||
                     (Type == ClassificationType.ClassifyBySize)) ?
                        true : false,
                    Coefficients,
                    new CFeaturesComparator[] { Comparator}, null, null),
                null, CFeatureMode.BuildArray(nbDim), true);
            result = opIns.Apply(analysis);

            if (Replace == true)
            {
                OperatorLevel1.CDelete opDel;
                opDel = new CDelete(EOCTypeNewEOC, EOCTypeChild, false, true);
                opDel.Apply(analysis);
            }

            Trace.Unindent();
            if (Trace.IndentLevel == 0) Trace.WriteLine("OK.", name);

            return result;
        }
        public override string ToString()
        {
            string features = ChildFeatures.ToString();
            string coefs = "{";
            foreach (double[] coef in Coefficients)
            {
                coefs += "{";
                foreach (double d in coef)
                {
                    coefs = coefs + d.ToString() + ",";
                }
                coefs += "},";
            }
            coefs += "}";
            string score = this.Comparator.ToString();
            return "Classify:" +
                "Type=" + System.Enum.GetName(typeof(ClassificationType), this.Type) + "," +
                "Parent=" + EOCTypeParent + "," +
                "Child=" + EOCTypeChild + "," +
                "New=" + EOCTypeNewEOC + "," +
                "Replace=" + Replace.ToString() + "," +
                "Features=" + ChildFeatures + "," +
                "Coefficients=" + coefs + "," +
                "Scores=" + score + "," +
                "";
        }
    }
}
