﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using Agora.Model.ImageAnalysis;
using Agora.Model.OperatorLevel0;
using Agora.Model.OperatorLevel0.Pairing;
using Agora.Model.OperatorLevel1;
using Agora.Model.OperatorLevel2;
using Agora.Model.Utils;

namespace Agora.Model.OperatorLevel3
{
    public class CEuclidianSpacing : Agora.Model.Project.CStepOperator
    {
        /// <summary>
        /// Context
        /// </summary>
        public string EOCTypeParent { get; set; }

        /// <summary>
        /// Type of the first child
        /// </summary>
        public string EOCTypeChild0 { get; set; }
        /// <summary>
        /// Type of the second child
        /// </summary>
        public string EOCTypeChild1 { get; set; }

        /// <summary>
        /// Type of the new child if classification has to be done
        /// </summary>
        public string EOCTypeNewEOC { get; set; }

        /// <summary>
        /// Direction H=true/V=false
        /// </summary>
        public bool Horizontal { get; set; }

        /// <summary>
        /// Classification type
        /// </summary>
        public enum SpacingType { Maximum, Minimum, Exact }
        public SpacingType Type { get; set; }

        /// <summary>
        /// Spacing XY features
        /// </summary>
        public double XSpacing { get; set; }
        public double XTolerance { get; set; }
        public double YSpacing { get; set; }
        public double YTolerance { get; set; }

        /// <summary>
        /// If new EOC is inserted, then replace child EOC
        /// instead of just insert it. This will call "Delete"
        /// </summary>
        public bool Replace { get; set; }

        public CEuclidianSpacing()
        {
            EOCTypeParent = null;
            EOCTypeChild0 = null;
            EOCTypeChild1 = null;
            EOCTypeNewEOC = null;
            Horizontal = true;
            Type = SpacingType.Exact;
            XSpacing = 0;
            XTolerance = 0;
            YSpacing = 0;
            YTolerance = 0;
            Replace = false;
        }

        public CEuclidianSpacing(
            string eocTypeParent, string eocTypeChild0, string eocTypeChild1, string eocTypeNewEOC,
            bool horizontal, SpacingType type,
            double xSpacing, double xTolerance, double yspacing, double yTolerance,
            bool replace
            )
        {
            EOCTypeParent = String.Copy(eocTypeParent);
            EOCTypeChild0 = String.Copy(eocTypeChild0);
            EOCTypeChild1 = String.Copy(eocTypeChild1);
            EOCTypeNewEOC = String.Copy(eocTypeNewEOC);
            Horizontal = horizontal;
            Type = type;
            XSpacing = xSpacing;
            XTolerance = xTolerance;
            YSpacing = yspacing;
            YTolerance = yTolerance;
            Replace = replace;
        }

        public override Agora.Model.Project.CStepOperator Clone()
        {
            return new CEuclidianSpacing(
                EOCTypeParent, EOCTypeChild0, EOCTypeChild1, EOCTypeNewEOC,
                Horizontal, Type,
                XSpacing, XTolerance, YSpacing, YTolerance,
                Replace);
        }

        public override bool Apply(CImageAnalysis analysis)
        {
            string name = new StackTrace().GetFrame(0).GetMethod().ReflectedType.Name;
            bool result = false;
            Trace.WriteLine("Euclidian spacing...", name);

            int nbDim = analysis.GetDimensions().Length;

            double[][] coefficients = CPairingEvaluator.CreateNullProjection(nbDim);
            int euclidian_X_num = analysis.GetDimensionIndex(CDimensionType.EUCLIDIAN_X);
            int euclidian_Y_num = analysis.GetDimensionIndex(CDimensionType.EUCLIDIAN_Y);
            coefficients[0][euclidian_X_num] = 1.0;
            coefficients[0][euclidian_Y_num] = 1.0;

            Trace.WriteLine("Phase 1 : Set Features...", name);
            Trace.Indent();

            CFeatureMode[] modes = null;
            OperatorLevel1.CSetFeatures opSetFeat = null;

            switch (Horizontal)
            {
                case true: // Mode Horizontal
                    modes = CFeatureMode.BuildArray(nbDim); 
                    modes[euclidian_X_num].aMode = CFeatureMode.SetFeatureMode.Min;
                    modes[euclidian_X_num].bMode = CFeatureMode.SetFeatureMode.Max;
                    modes[euclidian_Y_num].aMode = CFeatureMode.SetFeatureMode.AverageAll;
                    modes[euclidian_Y_num].bMode = CFeatureMode.SetFeatureMode.AverageAll;

                    opSetFeat = new CSetFeatures(EOCTypeParent, EOCTypeChild0, modes, true);
                    opSetFeat.Apply(analysis);
                    opSetFeat = new CSetFeatures(EOCTypeParent, EOCTypeChild1, modes, true);
                    opSetFeat.Apply(analysis);

                    break;
                case false: // Mode Vertical
                    modes = CFeatureMode.BuildArray(nbDim);
                    modes[euclidian_X_num].aMode = CFeatureMode.SetFeatureMode.AverageAll;
                    modes[euclidian_X_num].bMode = CFeatureMode.SetFeatureMode.AverageAll;
                    modes[euclidian_Y_num].aMode = CFeatureMode.SetFeatureMode.Min;
                    modes[euclidian_Y_num].bMode = CFeatureMode.SetFeatureMode.Max;

                    opSetFeat = new CSetFeatures(EOCTypeParent, EOCTypeChild0, modes, true);
                    opSetFeat.Apply(analysis);
                    opSetFeat = new CSetFeatures(EOCTypeParent, EOCTypeChild1, modes, true);
                    opSetFeat.Apply(analysis);
                    break;
            }

            Trace.Unindent();
            Trace.WriteLine("Phase 1 : OK.", name);

            Trace.WriteLine("Phase 2 : Insert " + EOCTypeNewEOC + "...", name);
            Trace.Indent();

            CFeaturesComparator comparator = new CFeaturesComparator(nbDim);
            switch (Type)
            {
                case SpacingType.Maximum:
                    switch (Horizontal)
                    {
                        case true:
                            comparator.Comparators[euclidian_X_num].a0b1.SignedDistance = true;
                            comparator.Comparators[euclidian_X_num].a0b1.MaxDistance = XSpacing;
                            comparator.Comparators[euclidian_Y_num].a0b1.SignedDistance = false;
                            comparator.Comparators[euclidian_Y_num].a0b1.MaxDistance = YTolerance;
                            break;
                        case false:
                            comparator.Comparators[euclidian_X_num].a0b1.SignedDistance = false;
                            comparator.Comparators[euclidian_X_num].a0b1.MaxDistance = YSpacing;
                            comparator.Comparators[euclidian_Y_num].a0b1.SignedDistance = true;
                            comparator.Comparators[euclidian_Y_num].a0b1.MaxDistance = XTolerance;
                            break;
                    }
                    break;
            }

            CFeaturesComparator[][] fc = CFeaturesComparator.BuildMatrix(2, nbDim);
            fc[0][1] = comparator;

            OperatorLevel1.CInsert opInsert = new CInsert(
                EOCTypeParent, new string[] { EOCTypeChild0, EOCTypeChild1 }, EOCTypeNewEOC,
                new CFeatures[] {
                    new CFeatures(new double[nbDim], new double[nbDim]),
                    new CFeatures(new double[nbDim], new double[nbDim]),
                    },
                new CPairingEvaluator(
                    false, false, false,
                    new double[][] { new double[nbDim] },
                    null, null, fc),
                null, CFeatureMode.BuildArray(nbDim), true);

            result = opInsert.Apply(analysis);

            Trace.Unindent();
            Trace.WriteLine("Phase 2 : OK.", name);
            Trace.WriteLine("Euclidian spacing OK.", name);

            return result;
        }

        public override string ToString()
        {
            return "EuclidianSpacing:" +
                "Parent=" + EOCTypeParent + "," +
                "Child0=" + EOCTypeChild0 + "," +
                "Child1=" + EOCTypeChild1 + "," +
                "New=" + EOCTypeNewEOC + "," +
                "H/V=" + Horizontal.ToString() + "," +
                "Type=" + Type.ToString() + "," +
                "XSpc=" + XSpacing.ToString() + "," +
                "XTol=" + XTolerance.ToString() + "," +
                "YSpc=" + XSpacing.ToString() + "," +
                "YTol=" + YTolerance.ToString() + "," +
                "Replace=" + Replace.ToString() + "," +
                "";
        }

    }
}
