﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using System.IO;
using System.Diagnostics;

using Agora.Model.ImageAnalysis;
using Agora.Model.OperatorLevel0;
using Agora.Model.Utils;

namespace Agora.Model.Project
{
    [XmlRoot("Scenario")]
    public class CScenario
    {
        [XmlIgnore]
        public bool NotSaved { get; set; }

        private CImageAnalysis _Analysis;
        [XmlIgnore]
        public CImageAnalysis Analysis
        {
            get { return _Analysis; }
        }

        private CDimensionType[] _Dimensions = null;
        [XmlIgnore]
        public CDimensionType[] Dimensions
        {
            get { return _Dimensions; }
            set { _Dimensions = value; }
        }

        private List<CScenarioStep> _Steps = new List<CScenarioStep>();
        [XmlArray("Steps")]
        [XmlArrayItem("Step", typeof(CScenarioStep))]
        public List<CScenarioStep> Steps
        {
            get { return _Steps; }
        }

        public CScenario()
        {
            NotSaved = false;
        }

        /// <summary>
        /// Prepare the scenario to be played, just before the 1st step.
        /// Init() doesn't clear the stored steps (you must explicitly call
        /// "ClearSteps()" if you want to start with a clean scenario)
        /// </summary>
        /// <param name="img"></param>
        public void Init(string img)
        {
            Trace.Assert(File.Exists(img));
            _Analysis = new CImageAnalysis();
            _Analysis.LoadImage(img);
            this.SetDimensions();
        }

        /// <summary>
        /// Set default dimensions (ie all)
        /// </summary>
        public void SetDimensions()
        {
            GLOBALDATA.DIMENSIONS = new CDimensionType[] {
                CDimensionType.EUCLIDIAN_X,
                CDimensionType.EUCLIDIAN_Y,
                CDimensionType.RANGEMAP_X,
                CDimensionType.RANGEMAP_Y,
                CDimensionType.CHILDS_NUMBER
            };
            this.SetDimensions(GLOBALDATA.DIMENSIONS);
        }
        private void SetDimensions(CDimensionType[] dimensions)
        {
            if (dimensions != null)
            {
                _Dimensions = (CDimensionType[])dimensions.Clone();
                _Analysis.SetDimensions(_Dimensions);
            }
        }

        public void ClearSteps()
        {
            while (_Steps.Count > 0)
            {
                this.RemoveLastStep();
            }
            _Steps.Clear();
            GLOBALDATA.ScenarioSteps.Clear();
        }

        public void AddStep(CScenarioStep step)
        {
            this.StepApply(step);
            _Steps.Add(step);
            NotSaved = true;
        }

        public void RemoveLastStep()
        {
            if (_Steps.Count > 0)
            {
                GLOBALDATA.ScenarioSteps.RemoveAt(GLOBALDATA.ScenarioSteps.Count - 1);
                CScenarioStep step = _Steps[_Steps.Count - 1];
                _Steps.RemoveAt(_Steps.Count - 1);
                this.StepUndo(step);
            }
            NotSaved = true;
        }

        public void StepApply(CScenarioStep step)
        {
            GLOBALDATA.ScenarioSteps.Add(step.ToString());
            step.SetAnalysis(_Analysis);
            _Analysis = step.Apply();
        }
        public void StepUndo(CScenarioStep step)
        {
            _Analysis = step.Undo();
        }

        /// <summary>
        /// Play ALL the stored steps
        /// </summary>
        public void Apply()
        {
            GLOBALDATA.ScenarioSteps.Clear();
            foreach (CScenarioStep step in _Steps)
            {
                this.StepApply(step);
            }
        }

        public void Save(string filename)
        {
            XmlSerializer xs = new XmlSerializer(typeof(CScenario));
            StreamWriter wr = new StreamWriter(filename);
            xs.Serialize(wr, this);
            wr.Close();
            NotSaved = false;
        }
        public static CScenario Load(string filename)
        {
            CScenario result;
            XmlSerializer xs = new XmlSerializer(typeof(CScenario));
            StreamReader rd = new StreamReader(filename);
            result = (CScenario)xs.Deserialize(rd);
            rd.Close();
            return result;
        }
    }
}
