﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MathNet.Numerics.LinearAlgebra;

namespace Agora.Model.Utils
{
    public class CLinearTransformation
    {
        public Matrix A = null;
        public Vector B = null;
        public CLinearTransformation()
        {
        }
        public CLinearTransformation(int n)
        {
            A = Matrix.Identity(n, n);
            B = new Vector(n, 0.0);
        }
        // copy constructor:
        public CLinearTransformation(CLinearTransformation obj)
        {
            this.A = obj.A.Clone();
            this.B = obj.B.Clone();
        }
        public Vector Transform(Vector U)
        {
            if (U == null)
                throw new Exception("Transform: U = null !");
            if ((A == null) || (B == null))
                throw new Exception("Transform: A or B = null !");
            if ((A.ColumnCount != U.Length) || (A.RowCount != U.Length) || (B.Length != U.Length))
                throw new Exception("Transform: sizes of A or B != size of U !");
            int size = U.Length;
            Vector V;
            V = A.Multiply(U);
            V.AddInplace(B);
            return V;
        }

        /// <summary>
        /// Compute transformation of x0 to y0e by y0e=A.x0+B
        /// using the transformation A,B given in parameter.
        /// A point y0e is computed for each x0 point.
        /// </summary>
        /// <param name="x0"></param>
        /// <param name="transformation"></param>
        /// <returns>List of vectors</returns>
        public List<Vector> TransformPoints(
            List<Vector> x0
            )
        {
            // Compute Y0Estim = A.X0 + B
            List<Vector> y0e = new List<Vector>();
            foreach (Vector x in x0)
            {
                Vector y = this.Transform(x);
                y0e.Add(y);
            }
            return y0e;
        }

        /// <summary>
        /// Finds and stores best transformation A,B for y0 = A.x0 + B
        /// </summary>
        /// <param name="x0">First set of points</param>
        /// <param name="y0">Second set of points</param>
        /// <param name="enableTranslation"></param>
        /// <param name="enableScaling"></param>
        /// <param name="enableRotation"></param>
        /// <returns></returns>
        public void ComputeBestTransformation(
            List<Vector> x0, List<Vector> y0,
            bool enableTranslation, bool enableScaling, bool enableRotation
            )
        {
            if (x0.Count != y0.Count)
            {
                throw new Exception("Lengths of point sets x0 and y0 are not identical");
            }
            if (enableTranslation == true)
            {
                // Compute averages of x0 and y0
                Vector x0moy = new Vector(x0[0].Length, 0.0);
                Vector y0moy = new Vector(y0[0].Length, 0.0);
                for (int i = 0; i < x0.Count; i++)
                {
                    x0moy.AddInplace(x0[i]);
                    y0moy.AddInplace(y0[i]);
                }
                x0moy.ScaleInplace(1.0 / (double)x0.Count);
                y0moy.ScaleInplace(1.0 / (double)y0.Count);

                // build x1 = list of centered pattern points (substract average)
                // build y1 = list of centered actual points (substract average)
                List<Vector> x1 = new List<Vector>();
                List<Vector> y1 = new List<Vector>();
                Vector v;
                for (int i = 0; i < x0.Count; i++)
                {
                    v = new Vector(x0[i].CopyToArray());
                    v.SubtractInplace(x0moy);
                    x1.Add(v);
                    v = new Vector(y0[i].CopyToArray());
                    v.SubtractInplace(y0moy);
                    y1.Add(v);
                }

                // find best adjustement Y0 = A.X0 + B
                // from previous step, we know average vectors /X0 and /Y0
                // 1. find best adjustement Y1 = A.X1
                //   a. Q = 1/n Y1.X1t (matrice covariance croisée centrée)
                //   b. I = 1/n X1.X1t (matrice covariance centrée)
                //   c. SVD of Q --> Q = LSV.S.RSV
                //   d. result 1 : scale factor s = Trace(S) / Trace(I)
                //   e. result 2 : rotation matrix R = LSV.RSV
                // 2. compute A, B
                //   a. A = sR
                //   b. B = /Y0 - A./X0

                double s;
                if (enableScaling == true)
                {
                    Matrix X1 = Matrix.CreateFromColumns(x1);
                    Matrix X1t = Matrix.CreateFromColumns(x1);
                    X1t.Transpose();
                    Matrix Y1 = Matrix.CreateFromColumns(y1);
                    Matrix Q = Y1.Multiply(X1t);
                    Q.Multiply(1.0 / (double)Y1.ColumnCount);
                    Matrix I = X1.Multiply(X1t);
                    I.Multiply(1.0 / (double)X1.ColumnCount);

                    if (enableRotation == true)
                    {
                        SingularValueDecomposition SVD = Q.SingularValueDecomposition;
                        Matrix LSV = SVD.LeftSingularVectors;
                        Matrix RSV = SVD.RightSingularVectors;
                        A = LSV.Multiply(RSV);
                        s = SVD.S.Trace() / I.Trace();
                    }
                    else
                    {
                        s = Q.Trace() / I.Trace();
                        A = Matrix.Identity(Q.RowCount, Q.RowCount);
                    }
                }
                else
                {
                    // no scaling, no rotation : A = identity
                    s = 1.0;
                    A = Matrix.Identity(x0[0].Length, x0[0].Length);
                }

                A.Multiply(s);
                B = A.Multiply(x0moy);
                B.ScaleInplace(-1.0);
                B.AddInplace(y0moy);
            }
            else
            {
                // No translation, no scaling, no rotation
                A = Matrix.Identity(x0[0].Length, x0[0].Length);
                B = new Vector(x0[0].Length); // 0.0 everywhere
            }
        }
    }
}
