﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using MathNet.Numerics.LinearAlgebra;

using Agora.Model.Utils;
using Agora.Model.ImageAnalysis;

namespace Agora.Model.Utils
{
    public class CNeighbourMapEntry
    {
        private Dictionary<CEOC, int> _IntervalNumber;
        private List<HashSet<CEOC>> _Intervals;
        private double _IntervalSize;

        public CNeighbourMapEntry(double intSize)
        {
            _IntervalSize = intSize;
            _Intervals = new List<HashSet<CEOC>>();
            _IntervalNumber = new Dictionary<CEOC, int>();
        }
        
        public void Remove(CEOC eoc)
        {
            /* old version without dictionary
            foreach (HashSet<CEOC> hs in this._Intervals)
            {
                hs.Remove(eoc);
            }
            */
            // new version with dictionary:
            if (_IntervalNumber.ContainsKey(eoc) == true)
            {
                _Intervals[_IntervalNumber[eoc]].Remove(eoc);
            }
        }

        public void Add(double val, CEOC eoc)
        {
            if (_Intervals.Count == 0)
            {
                _Intervals.Add(new HashSet<CEOC>());
            }
            if (val < 0.0)
            {
                // values between -inf and 0.0 --> first interval
                val = 0.0;
            }

            int numInterval = Convert.ToInt32(Math.Truncate(val / _IntervalSize));

            // Eventually add missing intervals
            for (int i = _Intervals.Count - 1; i < numInterval; i++)
            {
                _Intervals.Add(new HashSet<CEOC>());
            }

            // Store num in right interval
            _Intervals[numInterval].Add(eoc);
            // Store in dictionary for accelerating futur calls to GetNeighbors
            _IntervalNumber.Add(eoc, numInterval);
        }

        /// <summary>
        /// Retrieve neighbors of one EOC.
        /// A neighbor must :
        /// - be different from EOC
        /// - have the same parent than EOC
        /// - not be currently paired
        /// </summary>
        /// <param name="eoc">EOC</param>
        /// <param name="neighbors">Will contains found neighbors (no duplicate)</param>
        public void GetNeighbors(CEOC eoc, HashSet<CEOC> neighbors)
        {
            bool found = false;
            int numInterval = 0;

            // search interval which contains eoc
            found = _IntervalNumber.ContainsKey(eoc);
            if (found == true)
            {
                numInterval = _IntervalNumber[eoc];
            }

            // retrieve neighbors from interval-1 to interval+1
            if (found == true)
            {
                int minInterval = numInterval - 1;
                if (minInterval < 0)
                    minInterval = 0;
                int maxInterval = numInterval + 1;
                if (maxInterval >= _Intervals.Count)
                    maxInterval = _Intervals.Count - 1;
                for (int i = minInterval; i <= maxInterval; i++)
                {
                    foreach (CEOC neighbour in _Intervals[i])
                    {
                        if ((! neighbors.Contains(neighbour)) &&
                            (neighbour != eoc) &&
                            (neighbour.Parent == eoc.Parent) &&
                            (neighbour.PairedWith == null))
                        {
                            neighbors.Add(neighbour);
                        }
                    }
                }
            }
        }
    }

    public class CNeighbourMap
    {
        private List<CNeighbourMapEntry> _Map = null;
        public List<CNeighbourMapEntry> Map
        {
            get { return _Map; }
        }
        private int _EntryNumber;
        public int EntryNumber
        {
            get { return _EntryNumber; }
        }

        public CNeighbourMap(int entryNb)
        {
            _EntryNumber = entryNb;
            _Map = new List<CNeighbourMapEntry>(_EntryNumber);
        }

        public LinkedList<CEOC> GetNeighbors(CEOC eoc)
        {
            // create empty neighbors list
            LinkedList<CEOC> neighbors = new LinkedList<CEOC>();

            HashSet<CEOC> neighbourCandidates = new HashSet<CEOC>();
            HashSet<CEOC> neighbourCandidates2 = new HashSet<CEOC>();
            CNeighbourMapEntry mapEntry = _Map[0];
            mapEntry.GetNeighbors(eoc, neighbourCandidates);

            for (int numEntry = 1; numEntry < _Map.Count; numEntry++)
            {
                mapEntry = _Map[numEntry];
                neighbourCandidates2.Clear();
                mapEntry.GetNeighbors(eoc, neighbourCandidates2);
                neighbourCandidates.IntersectWith(neighbourCandidates2);
            }

            foreach (CEOC eocNeighbour in neighbourCandidates)
            {
                neighbors.AddLast(eocNeighbour);
            }
            return neighbors;
        }

        public void Builds(CEOC node, double[] neighbourhood)
        {
            for (int numCoord = 0; numCoord < neighbourhood.Length; numCoord++)
            {
                CNeighbourMapEntry nme = new CNeighbourMapEntry(neighbourhood[numCoord]);
                _Map.Add(nme);
            }

            LinkedListNode<CEOC> child = node.Childs.First;
            while (child != null)
            {
                this.AddNode(child.Value);
                child = child.Next;
            }
        }

        public void AddNode(CEOC node)
        {
            // compute average of points
            Vector v = new Vector((double[])node.Features.a.Clone());
            v.AddInplace(new Vector(node.Features.b));
            v.ScaleInplace(0.5);

            for (int numCoord = 0; numCoord < v.Length; numCoord++)
            {
                _Map[numCoord].Add(v[numCoord], node);
            }
        }

        public void RemoveNode(CEOC node)
        {
            CNeighbourMapEntry mapEntry;
            for (int numEntry = 0; numEntry < _Map.Count; numEntry++)
            {
                mapEntry = _Map[numEntry];
                mapEntry.Remove(node);
            }
        }
    }
}
