
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Security.Cryptography;

namespace Agora.Model.Utils
{
    unsafe public class FastBitmap
    {
        public struct PixelData32bpp
        {
            public byte blue;
            public byte green;
            public byte red;
            public byte alpha;

            public override string ToString()
            {
                return "(" + alpha.ToString() + ", " + red.ToString() + ", " + green.ToString() + ", " + blue.ToString() + ")";
            }
        }

        public struct PixelData24bpp
        {
            public byte blue;
            public byte green;
            public byte red;

            public override string ToString()
            {
                return "(" + red.ToString() + ", " + green.ToString() + ", " + blue.ToString() + ")";
            }
        }

        public struct PixelData8bpp
        {
            public byte gray;

            public override string ToString()
            {
                return "(" + gray.ToString() + ")";
            }
        }

        private Bitmap _WorkingBitmap = null;
        private int _Height = 0;
        private int _Width = 0;
        private BitmapData _BitmapData = null;
        private Byte* _PtrBase = null;

        public FastBitmap(Bitmap inputBitmap)
        {
            if ((inputBitmap.PixelFormat == PixelFormat.Format32bppArgb) ||
                (inputBitmap.PixelFormat == PixelFormat.Format24bppRgb) ||
                (inputBitmap.PixelFormat == PixelFormat.Format8bppIndexed))
            {
                _WorkingBitmap = inputBitmap;
            }
            else
            {
                throw new Exception("FastBitmap supports only 8bpp, 24bpp and 32bpp images");
            }
        }

        public void LockImage()
        {
            Rectangle bounds = new Rectangle(Point.Empty, _WorkingBitmap.Size);

            //Lock Image
            _BitmapData = _WorkingBitmap.LockBits(bounds, ImageLockMode.ReadWrite, _WorkingBitmap.PixelFormat);
            _Width = _BitmapData.Stride;
            _Height = _BitmapData.Height;
            _PtrBase = (Byte*)_BitmapData.Scan0.ToPointer();
        }
        public void UnlockImage()
        {
            _WorkingBitmap.UnlockBits(_BitmapData);
            _BitmapData = null;
            _PtrBase = null;
        }

        public Color GetPixel(int x, int y)
        {
            Color result = Color.Black;
            PixelData32bpp* pixelData32bpp = null;
            PixelData24bpp* pixelData24bpp = null;
            PixelData8bpp* pixelData8bpp = null;
            switch (_WorkingBitmap.PixelFormat)
            {
                case PixelFormat.Format32bppArgb:
                    pixelData32bpp = (PixelData32bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData32bpp));
                    result = Color.FromArgb(pixelData32bpp->alpha, pixelData32bpp->red, pixelData32bpp->green, pixelData32bpp->blue);
                    break;
                case PixelFormat.Format24bppRgb:
                    pixelData24bpp = (PixelData24bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData24bpp));
                    result = Color.FromArgb(255, pixelData24bpp->red, pixelData24bpp->green, pixelData24bpp->blue);
                    break;
                case PixelFormat.Format8bppIndexed:
                    pixelData8bpp = (PixelData8bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData8bpp));
                    result = Color.FromArgb(255, pixelData8bpp->gray, pixelData8bpp->gray, pixelData8bpp->gray);
                    break;
            }
            return result;
        }
        public PixelData32bpp GetPixel32bpp(int x, int y)
        {
            PixelData32bpp* pixelData32bpp = (PixelData32bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData32bpp));
            return *pixelData32bpp;
        }
        public PixelData24bpp GetPixel24bpp(int x, int y)
        {
            PixelData24bpp* pixelData24bpp = (PixelData24bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData24bpp));
            return *pixelData24bpp;
        }
        public PixelData8bpp GetPixel8bpp(int x, int y)
        {
            PixelData8bpp* pixelData8bpp = (PixelData8bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData8bpp));
            return *pixelData8bpp;
        }

        public void SetPixel(int x, int y, Color color)
        {
            PixelData32bpp* pixelData32bpp = null;
            PixelData24bpp* pixelData24bpp = null;
            PixelData8bpp* pixelData8bpp = null;
            switch (_WorkingBitmap.PixelFormat)
            {
                case PixelFormat.Format32bppArgb:
                    pixelData32bpp = (PixelData32bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData32bpp));
                    pixelData32bpp->alpha = color.A;
                    pixelData32bpp->red = color.R;
                    pixelData32bpp->green = color.G;
                    pixelData32bpp->blue = color.B;
                    break;
                case PixelFormat.Format24bppRgb:
                    pixelData24bpp = (PixelData24bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData24bpp));
                    pixelData24bpp->red = color.R;
                    pixelData24bpp->green = color.G;
                    pixelData24bpp->blue = color.B;
                    break;
                case PixelFormat.Format8bppIndexed:
                    pixelData8bpp = (PixelData8bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData8bpp));
                    pixelData8bpp->gray = color.R;
                    break;
            }
        }
        public void SetPixel32bpp(int x, int y, PixelData32bpp pix)
        {
            PixelData32bpp* pixelData32bpp = (PixelData32bpp*)(_PtrBase + y * _Width + x * sizeof(PixelData32bpp));
            pixelData32bpp->alpha = pix.alpha;
            pixelData32bpp->red = pix.red;
            pixelData32bpp->green = pix.green;
            pixelData32bpp->blue = pix.blue;
        }

        public new byte[] GetHashCode()
        {
            byte[] result = null;

            int byteCount = _Width * _Height;
            byte[] bmpBytes = new byte[byteCount];

            // Copy the locked bytes from memory
            Marshal.Copy(_BitmapData.Scan0, bmpBytes, 0, byteCount);

            MD5 md5Hasher = MD5.Create();
            result = md5Hasher.ComputeHash(bmpBytes);

            return result;
        }

        public unsafe Bitmap Clone(int startX, int startY, int lengthX, int lengthY)
        {
            Bitmap croppedBitmap = new Bitmap(lengthX, lengthY, _WorkingBitmap.PixelFormat);
            FastBitmap fastCroppedBitmap = new FastBitmap(croppedBitmap);
            fastCroppedBitmap.LockImage();

            int BPP = 0;
            switch (_WorkingBitmap.PixelFormat)
            {
                case PixelFormat.Format32bppArgb: BPP = sizeof(PixelData32bpp); break;
                case PixelFormat.Format24bppRgb: BPP = sizeof(PixelData24bpp); break;
                case PixelFormat.Format8bppIndexed: BPP = sizeof(PixelData8bpp); break;
            }

            //Iterate the selected area of the original image, and the full area of the new image 
            for (int y = 0; y < lengthY; y++)
            {
                Byte* pOriginal = this._PtrBase + (startY + y) * this._Width + startX * BPP;
                Byte* pCropped = fastCroppedBitmap._PtrBase + y * fastCroppedBitmap._Width;
                for (int x = 0; x < lengthX * BPP; x++)
                {
                    *pCropped++ = *pOriginal++;
                }
            }

            fastCroppedBitmap.UnlockImage();

            return croppedBitmap;
        }
    }
}
