﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using Agora.Model.ImageAnalysis;
using Agora.Model.Project;

namespace Agora.ViewModel
{
    public class CExceptionViewModel : System.Exception
    {
        public enum Type
        {
            ProjectNotExisting = 0,
            ProjectNotSaved,
            ProjectCannotLoad,
            ProjectCannotSave,
            ProjectSubDirectoryNotInProjectDirectory,
            ProjectCurrentImageNotExisting,
            ScenarioNotExisting,
            ScenarioNotSaved,
            ScenarioCannotLoad,
            ScenarioCannotSave,
            ScenarioDimensionsNotSet,
        }
        public static string[] Description =
        {
            "Project not existing",
            "Project not saved",
            "Cannot load project",
            "Cannot save project",
            "Cannot set subdirectory : not in project's directory",
            "Cannot set current image",
            "Scenario not existing",
            "Scenario not saved",
            "Cannot load scenario",
            "Cannot save scenario",
            "Dimensions not set",
        };

        public CExceptionViewModel(CExceptionViewModel.Type t)
            : base(Description[(int)t])
        {
            
        }
        public CExceptionViewModel(CExceptionViewModel.Type t, string m)
            : base(Description[(int)t] + ": " + m)
        {

        }
    }

    /// <summary>
    /// This class is responsible to provide GUI-friendly access
    /// to data in the Agora's model ("Analysis status"). The model
    /// is strictly private, and can only be accessed by using methods
    /// of the CViewModel class.
    /// Moreover, this class provides methods to launch all the methods
    /// of Agora's objects ("XXX Operations").
    /// </summary>
    public class CViewModel
    {
        private CProject _Project = new CProject();

        #region Project operations

        public void ProjectCreateNew()
        {
            this.ProjectClose();
            _Project = new CProject();
        }
        public void ProjectSave(string filename)
        {
            try
            {
                _Project.Save(filename);
            }
            catch (Exception ex)
            {
                throw new CExceptionViewModel(
                    CExceptionViewModel.Type.ProjectCannotSave,
                    filename + " (" + ex.Message + ")");
            }
        }
        public void ProjectOpen(string filename)
        {
            try
            {
                this.ProjectClose();
            }
            catch (Exception ex)
            {
            }
            //try
            //{
                _Project = CProject.Load(filename);
            //}
            /*
             * catch (Exception ex)
            {
                throw new CExceptionViewModel(
                    CExceptionViewModel.Type.ProjectCannotLoad,
                    filename + " (" + ex.Message + ")" );
            }
             */
        }
        public void ProjectClose()
        {
            if (_Project.NotSaved == true)
            {
                throw new CExceptionViewModel(CExceptionViewModel.Type.ProjectNotSaved);
            }
            if (_Project.Scenario.NotSaved == true)
            {
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioNotSaved);
            }
            _Project = null;
        }

        public void ProjectSetProjectDirectory(string dir)
        {
            _Project.SetDirectory(dir);
        }
        public string ProjectGetProjectDirectory()
        {
            return _Project.GetDirectory();
        }
        public void ProjectSetImagesSubDirectory(string dir)
        {
            if (dir.StartsWith(CProject.Directory) == false)
            {
                throw new CExceptionViewModel(CExceptionViewModel.Type.ProjectSubDirectoryNotInProjectDirectory);
            }
            string[] subdirs = dir.Split(System.IO.Path.DirectorySeparatorChar);
            string imagesSubdir = subdirs[subdirs.Length - 1];
            _Project.SetImagesSubDirectory(imagesSubdir);
        }
        public string ProjectGetImagesSubDirectory()
        {
            return _Project.GetImagesSubDirectory();
        }

        public void ProjectSelectCurrentImage(string filename)
        {
            if (_Project.ImageFilenames.Contains(filename) == false)
            {
                throw new CExceptionViewModel(CExceptionViewModel.Type.ProjectCurrentImageNotExisting);
            }
            _Project.SetCurrentImage(filename);
        }
        public List<string> ProjectGetImages()
        {
            return _Project.ImageFilenames;
        }
        public void ProjectScenarioLoad(string filename)
        {
            try
            {
                _Project.ScenarioLoad(filename);
            }
            catch (Exception ex)
            {
                throw new CExceptionViewModel(
                    CExceptionViewModel.Type.ScenarioCannotLoad,
                    filename + " (" + ex.Message + ")");
            }
        }
        public void ProjectScenarioSave(string filename)
        {
            try
            {
                _Project.ScenarioSave(filename);
            }
            catch (Exception ex)
            {
                throw new CExceptionViewModel(
                    CExceptionViewModel.Type.ScenarioCannotSave,
                    filename + " (" + ex.Message + ")");
            }
        }
        public void ProjectScenarioInit()
        {
            _Project.ScenarioInit();
        }
        public void ProjectScenarioClear()
        {
            _Project.ScenarioClear();
        }
        public void ProjectApply()
        {
            _Project.Apply();
        }
        #endregion

        #region Scenario operations

        public void ScenarioSetDimensions()
        {
            _Project.Scenario.SetDimensions();
        }

        public List<string> ScenarioGetStepStrings()
        {
            List<string> result = new List<string>();
            foreach (CScenarioStep step in _Project.Scenario.Steps)
            {
                string s = step.ToString();
                result.Add(s);
            }
            return result;
        }

        /// <summary>
        /// Returns a list of CLONES of operator in the scenario
        /// </summary>
        /// <returns></returns>
        public List<CStepOperator> ScenarioGetStepOperators()
        {
            List<CStepOperator> operators = new List<CStepOperator>();
            foreach (CScenarioStep step in _Project.Scenario.Steps)
            {
                CStepOperator op = step.Operator.Clone();
                operators.Add(op);
            }
            return operators;
        }

        public void ScenarioRemoveLastStep()
        {
            _Project.Scenario.RemoveLastStep();
        }

        #endregion

        #region L1 operations

        private Agora.Model.OperatorLevel1.CDelete _L1OperatorDelete =
            new Agora.Model.OperatorLevel1.CDelete();
        public Agora.Model.OperatorLevel1.CDelete L1OperatorDelete
        {
            get { return _L1OperatorDelete; }
        }
        public void L1Delete()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L1OperatorDelete.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel1.CFindCC _L1OperatorFindCC =
            new Agora.Model.OperatorLevel1.CFindCC();
        public Agora.Model.OperatorLevel1.CFindCC L1OperatorFindCC
        {
            get { return _L1OperatorFindCC; }
        }
        public void L1FindCC()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L1OperatorFindCC.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel1.CInsert _L1OperatorInsert =
            new Agora.Model.OperatorLevel1.CInsert();
        public Agora.Model.OperatorLevel1.CInsert L1OperatorInsert
        {
            get { return _L1OperatorInsert; }
        }
        public void L1Insert()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L1OperatorInsert.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel1.CSetFeatures _L1OperatorSetFeatures =
            new Agora.Model.OperatorLevel1.CSetFeatures();
        public Agora.Model.OperatorLevel1.CSetFeatures L1OperatorSetFeatures
        {
            get { return _L1OperatorSetFeatures; }
        }
        public void L1SetFeatures()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);

            CStepOperator op = L1OperatorSetFeatures.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        #endregion

        #region L2 Operations

        private Agora.Model.OperatorLevel2.CClassifyByFeatures _L2OperatorClassifyByFeatures =
            new Agora.Model.OperatorLevel2.CClassifyByFeatures();
        public Agora.Model.OperatorLevel2.CClassifyByFeatures L2OperatorClassifyByFeatures
        {
            get { return _L2OperatorClassifyByFeatures; }
        }
        public void L2ClassifyByFeatures()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L2OperatorClassifyByFeatures.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel2.CFusion _L2OperatorFusion =
            new Agora.Model.OperatorLevel2.CFusion();
        public Agora.Model.OperatorLevel2.CFusion L2OperatorFusion
        {
            get { return _L2OperatorFusion; }
        }
        public void L2Fusion()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L2OperatorFusion.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel2.CIntersect _L2OperatorIntersect =
            new Agora.Model.OperatorLevel2.CIntersect();
        public Agora.Model.OperatorLevel2.CIntersect L2OperatorIntersect
        {
            get { return _L2OperatorIntersect; }
        }
        public void L2Intersect()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L2OperatorIntersect.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        #endregion

        #region L3 Operations

        private Agora.Model.OperatorLevel3.CTextSegmentation _L3OperatorTextSegmentation =
            new Agora.Model.OperatorLevel3.CTextSegmentation();
        public Agora.Model.OperatorLevel3.CTextSegmentation L3OperatorTextSegmentation
        {
            get { return _L3OperatorTextSegmentation; }
        }
        public void L3TextSegmentation()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L3OperatorTextSegmentation.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel3.CEuclidianFeatures _L3OperatorEuclidianFeatures =
            new Agora.Model.OperatorLevel3.CEuclidianFeatures();
        public Agora.Model.OperatorLevel3.CEuclidianFeatures L3OperatorEuclidianFeatures
        {
            get { return _L3OperatorEuclidianFeatures; }
        }
        public void L3EuclidianFeatures()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L3OperatorEuclidianFeatures.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorLevel3.CEuclidianSpacing _L3OperatorEuclidianSpacing =
            new Agora.Model.OperatorLevel3.CEuclidianSpacing();
        public Agora.Model.OperatorLevel3.CEuclidianSpacing L3OperatorEuclidianSpacing
        {
            get { return _L3OperatorEuclidianSpacing; }
        }
        public void L3EuclidianSpacing()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = L3OperatorEuclidianSpacing.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        #endregion

        #region Misc operations

        private Agora.Model.OperatorMisc.CCreateLabel _MiscOperatorCreateLabel =
            new Agora.Model.OperatorMisc.CCreateLabel();
        public Agora.Model.OperatorMisc.CCreateLabel MiscOperatorCreateLabel
        {
            get { return _MiscOperatorCreateLabel; }
        }
        public void MiscCreateLabel()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = MiscOperatorCreateLabel.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorMisc.CDeleteLabel _MiscOperatorDeleteLabel =
            new Agora.Model.OperatorMisc.CDeleteLabel();
        public Agora.Model.OperatorMisc.CDeleteLabel MiscOperatorDeleteLabel
        {
            get { return _MiscOperatorDeleteLabel; }
        }
        public void MiscDeleteLabel()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = MiscOperatorDeleteLabel.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        private Agora.Model.OperatorMisc.CExportResults _MiscOperatorExportResults =
            new Agora.Model.OperatorMisc.CExportResults();
        public Agora.Model.OperatorMisc.CExportResults MiscOperatorExportResults
        {
            get { return _MiscOperatorExportResults; }
        }
        public void MiscExportResults()
        {
            if (_Project.Scenario.Dimensions == null)
                throw new CExceptionViewModel(CExceptionViewModel.Type.ScenarioDimensionsNotSet);
            CStepOperator op = MiscOperatorExportResults.Clone();
            CScenarioStep step = new CScenarioStep(op);
            _Project.Scenario.AddStep(step);
        }

        #endregion

        #region Analysis status

        public CDimensionType[] AnalysisGetDimensions()
        {
            CDimensionType[] dimensions = _Project.Scenario.Analysis.GetDimensions();
            return dimensions;
        }

        public List<string> AnalysisGetLabels()
        {
            List<string> labels = _Project.Scenario.Analysis.GetEOCTypeNames();
            return labels;
        }

        public System.Drawing.Rectangle AnalysisGraphGetBoundingBox(Object obj)
        {
            System.Drawing.Rectangle result = new System.Drawing.Rectangle();
            if (obj.GetType() == typeof(CEOC))
            {
                CEOC eoc = (CEOC)obj;
                result = eoc.BoundingBox;
            }
            return result;
        }

        public List<System.Drawing.Rectangle> AnalysisGetBoundingBoxes(string label)
        {
            List<System.Drawing.Rectangle> result = new List<System.Drawing.Rectangle>();
            CEOCType eocType = _Project.Scenario.Analysis.GetEOCTypeByName(label);
            CEOC eoc;
            Agora.Model.Utils.CBFS bfs = new Agora.Model.Utils.CBFS();
            bfs.Start(_Project.Scenario.Analysis.Graph.GetRoot());
            while ((eoc = bfs.GetNext()) != null)
            {
                if (eoc.Eoctype == eocType)
                {
                    result.Add(eoc.BoundingBox);
                }
            }
            return result;
        }

        public void AnalysisGenerateDebugImage(string filename, List<string> labels)
        {
            System.Drawing.KnownColor[] labelColors =
            {
                System.Drawing.KnownColor.Red,
                System.Drawing.KnownColor.Yellow,
                System.Drawing.KnownColor.Green,
                System.Drawing.KnownColor.Orange,
                System.Drawing.KnownColor.Blue,
            };
            string curImageFilename = _Project.ImageFilenames[_Project.CurrentImageNum];
            System.Drawing.Bitmap image = new System.Drawing.Bitmap(_Project.Scenario.Analysis.CurrentImage);
            System.Drawing.Graphics graphics = System.Drawing.Graphics.FromImage(image);
            foreach (string label in labels)
            {
                int labelNum = labels.IndexOf(label);
                System.Drawing.Pen pen = new System.Drawing.Pen(System.Drawing.Color.FromKnownColor(labelColors[labelNum]));
                pen.Width = 1.0F;
                List<System.Drawing.Rectangle> boundingBoxes = this.AnalysisGetBoundingBoxes(label);
                foreach (System.Drawing.Rectangle rect in boundingBoxes)
                {
                    System.Drawing.Rectangle r = new System.Drawing.Rectangle(
                        rect.X, rect.Y, rect.Width - 1, rect.Height - 1);
                    graphics.DrawRectangle(pen, r);
                }
                pen.Dispose();
            }
            graphics.Dispose();
            image.Save(filename);
            image.Dispose();
        }

        public System.Drawing.Bitmap AnalysisGetCurrentImage()
        {
            System.Drawing.Bitmap bitmap = _Project.Scenario.Analysis.CurrentImage;
            return bitmap;
        }

        public TreeNode SearchNearestTreeNode(TreeNode root, TreeNode bestfound, string label, int x, int y)
        {
            foreach (TreeNode child in root.Nodes)
            {
                bestfound = this.SearchNearestTreeNode(child, bestfound, label, x, y);
                CEOC eoc = (CEOC)child.Tag;
                if (eoc.Eoctype.TypeName == label)
                {
                    CEOC eocBestFound = (CEOC)bestfound.Tag;
                    double dx = (double)x - (double)eocBestFound.BoundingBox.X;
                    double dy = (double)y - (double)eocBestFound.BoundingBox.Y;
                    double distanceBestFound = Math.Sqrt(dx * dx + dy * dy);
                    dx = (double)x - (double)eoc.BoundingBox.X;
                    dy = (double)y - (double)eoc.BoundingBox.Y;
                    double distance = Math.Sqrt(dx * dx + dy * dy);
                    if (distance < distanceBestFound)
                    {
                        distanceBestFound = distance;
                        bestfound = child;
                    }
                }
            }
            return bestfound;
        }
        public TreeNode AnalysisGetGraph()
        {
            TreeNode root = this.GetTreeNode(_Project.Scenario.Analysis.Graph.GetRoot());
            return root;
        }
        private TreeNode GetTreeNode(CEOC eoc)
        {
            TreeNode node = new TreeNode();
            List<string> labels = this.AnalysisGetLabels();
            string features = eoc.Features.ToString();
            node.Text = eoc.Eoctype.TypeName + eoc.BoundingBox.ToString() + features;
            node.ImageIndex = labels.IndexOf(eoc.Eoctype.TypeName);
            node.Tag = eoc;
            foreach (CEOC childEOC in eoc.Childs)
            {
                TreeNode childNode = this.GetTreeNode(childEOC);
                node.Nodes.Add(childNode);
            }
            return node;
        }
        #endregion
    }
}
